# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import os
import pathlib
import subprocess
import tempfile

import networkx as nx
import pulp
import tabulate

from slambuc.alg import LP_LAT, T_PART, T_BARRS
from slambuc.alg.service.common import *
from slambuc.alg.util import (ichain, path_blocks, block_memory, block_cost, block_latency, leaf_label_nodes,
                              block_cpu, ser_block_subcost, ser_block_sublatency, ser_block_submemory,
                              ser_subtree_cost, ser_subchain_latency, par_subtree_cost, par_subchain_latency)
from slambuc.misc.plot import draw_tree


def get_cplex_path() -> str:
    """
    Return local CPLEX path.

    :return: path
    """
    if cplex_path := subprocess.run(['which', 'cplex'], stdout=subprocess.PIPE).stdout.decode().strip():
        return cplex_path
    cplex_path = pathlib.Path(os.environ.get('CPLEX_HOME', '~/Programs/ibm/ILOG/CPLEX_Studio2211/cplex'),
                              'bin/x86-64_linux/cplex').expanduser()
    return str(cplex_path) if cplex_path.exists() else None


def get_cpo_path() -> str:
    """
    Return local CPO path.

    :return: path
    """
    if cpo_path := subprocess.run(['which', 'cpoptimizer'], stdout=subprocess.PIPE).stdout.decode().strip():
        return cpo_path
    cpo_path = pathlib.Path(os.environ.get('CPO_HOME', '~/Programs/ibm/ILOG/CPLEX_Studio2211/cpoptimizer'),
                            'bin/x86-64_linux/cpoptimizer').expanduser()
    return str(cpo_path) if cpo_path.exists() else None


def get_glpk_path() -> str:
    """
    Return local GLPK path.

    :return: path
    """
    if glpk_path := subprocess.run(['which', 'glpsol'], stdout=subprocess.PIPE).stdout.decode().strip():
        return glpk_path
    else:
        return None


def is_compatible(tree1: nx.DiGraph, tree2: nx.DiGraph) -> bool:
    """
    Return true if given second *tree2* has the same structure and edge/node attributes as the first *tree1*.

    :param tree1:   first tree
    :param tree2:   second tree
    :return:        similarity result
    """
    return (nx.is_isomorphic(tree1, tree2) and
            all(all(tree1[u][v][a] == tree2[u][v][a] for a in tree1[u][v]) and
                all(tree1.nodes[v][a] == tree2.nodes[v][a] for a in tree1.nodes[v]) for u, v in tree1.edges))


def get_chain_k_min(memory: list[int], M: int, rate: list[int], N: int, start: int = 0, end: int = None) -> int:
    """
    Return minimal number of blocks due to constraints *M* and *N*.

    :param memory:  list of memory values
    :param M:       memory upper bound
    :param rate:    list of rate values
    :param N:       CPU count
    :param start:   fist node to consider
    :param end:     last node to consider
    :return:        minimal number of blocks
    """
    end = end if end is not None else len(memory) - 1
    return max(math.ceil(block_memory(memory, start, end) / M),
               sum(1 for i, j in itertools.pairwise(rate[start: end + 1]) if math.ceil(j / i) > N))


def get_chain_c_min(memory: list[int], M: int, rate: list[int], N: int, start: int = 0, end: int = None) -> int:
    """
    Return minimal number of cuts due to constraints *M* and *N*.

    :param memory:  list of memory values
    :param M:       memory upper bound
    :param rate:    list of rate values
    :param N:       CPU count
    :param start:   fist node to consider
    :param end:     last node to consider
    :return:        minimal number of cuts
    """
    return get_chain_k_min(memory, M, rate, N, start, end) - 1


def get_chain_c_max(runtime: list[int], L: int, b: int, w: int, delay: int, start: int = 0, end: int = None) -> int:
    """
    Return maximal number of cuts due to constraint *L*.

    :param runtime: list of runtime values
    :param L:       upper latency limit
    :param b:       barrier node
    :param w:       end node of chain block
    :param delay:   platform delay
    :param start:   fist node to consider
    :param end:     last node to consider
    :return:        maximum number of cuts
    """
    end = end if end is not None else len(runtime) - 1
    return math.floor(min((L - block_latency(runtime, b, w, delay, start, end)) / delay, len(runtime) - 1))


def get_chain_k_max(runtime: list[int], L: int, b: int, w: int, delay: int, start: int = 0, end: int = None) -> int:
    """
    Return maximal number of blocks due to constraint *L*.

    :param runtime: list of runtime values
    :param L:       upper latency limit
    :param b:       barrier node
    :param w:       end node of chain block
    :param delay:   platform delay
    :param start:   fist node to consider
    :param end:     last node to consider
    :return:        maximum number of blocks
    """
    return get_chain_c_max(runtime, L, b, w, delay, start, end) + 1


def get_chain_k_opt(partition: T_PART, start: int = 0, end: int = None) -> int:
    """
    Return the number of blocks included by the [*start*, *end*] interval in partitioning.

    :param partition:   chain partitioning
    :param start:       fist node to consider
    :param end:         last node to consider
    :return:            number of blocks
    """
    end = end if end is not None else partition[-1][-1]
    cntr = 0
    in_chain = False
    for blk in partition:
        b, w = blk[0], blk[-1]
        if not in_chain and b <= start:
            in_chain = True
        if in_chain:
            cntr += 1
        if in_chain and end <= w:
            in_chain = False
    return cntr


def get_chain_c_opt(partition: T_PART, start: int = 0, end: int = None) -> int:
    """
    Return the number of cuts included by the [*start*, *end*] interval in partitioning.

    :param partition:   chain partitioning
    :param start:       fist node to consider
    :param end:         last node to consider
    :return:            number of cuts
    """
    return get_chain_k_opt(partition, start, end) - 1


def prune_chain(tree: nx.DiGraph, node: int, leaf: int) -> tuple[list[int], list[int]]:
    """
    Return the nodes of chain [*node*, *leaf*] and the branching nodes.

    :param tree:    service tree
    :param node:    chain's barrier node
    :param leaf:    end node of chain
    :return:        nodes of the chain and its branches
    """
    chain = [node]
    branches = []
    u = node
    while u != leaf:
        for _, v in tree.out_edges(u):
            if leaf in tree.nodes[v][LABEL]:
                chain.append(v)
            else:
                branches.append(v)
        u = chain[-1]
    return chain, branches


########################################################################################################################


def print_chain_summary(runtime: list[int], memory: list[int], rate: list[int]):
    """
    Print chain summary.

    :param runtime: list of runtime values
    :param memory:  list of memory values
    :param rate:    list of rate values
    """
    print("Chain:", "[", *(f"-{r}-> F({t}|M{m})" for t, m, r in zip(runtime, memory, rate)), "]")


def evaluate_chain_partitioning(partition: T_PART, opt_cost: int, opt_lat: int, runtime: list, memory: list, rate: list,
                                M: int = math.inf, N: int = math.inf, L: int = math.inf, start: int = 0,
                                end: int = None, delay: int = 1, unit: int = 100):
    """
    Evaluate chain partitioning and print its characteristics.

    :param partition:   chain partitioning
    :param opt_cost:    optimal cost of the partitioning
    :param opt_lat:     latency of the partitioning
    :param runtime:     list of runtime values
    :param memory:      list of memory values
    :param rate:        list of rate values
    :param M:           memory upper bound
    :param N:           CPU count
    :param L:           upper latency limit
    :param start:       fist node to consider
    :param end:         last node to consider
    :param delay:       platform delay
    :param unit:        rounding unit
    """
    print('#' * 80)
    print(f"Chain partitioning [M={M}, N={N}, L={L}:{(start, end)}] => "
          f"{partition} - opt_cost: {opt_cost}, opt_lat: {opt_lat}")
    print(f"k_min={get_chain_k_min(memory, M, rate, N, start, end)}, "
          f"k_opt[L]={len(path_blocks(partition, range(start, end + 1)))}, "
          f"k_max={get_chain_k_max(runtime, L, 0, len(runtime), delay, start, end)}")
    print_block_stat(partition, runtime, memory, rate, delay, start, end, unit)
    print('#' * 80)


def print_block_stat(partition: T_PART, runtime: list[int], memory: list[int], rate: list[int], delay: float,
                     start: int = 0, end: int = None, unit: int = 100):
    """
    Print block statistics.

    :param partition:   chain partitioning
    :param runtime:     list of runtime values
    :param memory:      list of memory values
    :param rate:        list of rate values
    :param start:       fist node to consider
    :param end:         last node to consider
    :param delay:       platform delay
    :param unit:        rounding unit
    """
    end = end if end is not None else len(runtime) - 1
    stat = [[str([blk[0], blk[-1]]),
             block_cost(runtime, rate, blk[0], blk[-1], unit),
             block_memory(memory, blk[0], blk[-1]),
             block_cpu(rate, blk[0], blk[-1]),
             block_latency(runtime, blk[0], blk[-1], delay, start, end)] for blk in partition]
    print(tabulate.tabulate(stat, ['Block', 'Cost', 'Memory', 'CPU', 'Latency'],
                            numalign='decimal', stralign='center', tablefmt='pretty'))


def print_chain_partition_result(barr: T_BARRS, cost: int, lat: int):
    """
    Decode and print chain partitioning result.

    :param barr:    barrier nodes
    :param cost:    optimal cost
    :param lat:     latency values
    """
    if barr is not None:  # [], cost, lat
        print("Minimal-cost solution is calculated.")
    elif cost is not None and lat is None:  # None, inf, None
        print("Feasible latency-fitting solution cannot be generated due to memory constraint M.")
    elif cost is None and lat is not None:  # None, None, lat
        print("Solution does not exist due to too strict latency constraint L.")
    else:  # None, None, None
        print("Feasible solution does not exist due to non-overlapping constrains L and M.")


def print_tree_summary(tree: nx.DiGraph):
    """
    Print summary of service graphs.

    :param tree:    input tree
    """
    print(tree)
    for n, nd in tree.nodes(data=True):
        print(f"\t{n}: {nd}")
        for i, j, ed in tree.out_edges(n, data=True):
            print(f"\t\t{i} -> {j}: {ed}")


def print_tree_block_stat(tree: nx.DiGraph, partition: T_PART, unit: int = 100):
    """
    Print cost memory and latency values of partition blocks in tabulated format.

    :param tree:        input tree
    :param partition:   given partitioning
    :param unit:        rounding unit
    """
    stat = []
    for blk in partition:
        pred = next(tree.predecessors(blk[0]))
        runtime, memory, rate = zip(*[(tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY], tree[u][v][RATE])
                                      for u, v in itertools.pairwise([pred] + blk)])
        b, w = 0, len(blk) - 1
        stat.append([str([blk[b], blk[w]]),
                     block_cost(runtime, rate, b, w, unit),
                     block_memory(memory, b, w),
                     block_cpu(rate, b, w),
                     block_latency(runtime, b, w, 0, b, w)])
    print(tabulate.tabulate(stat, ['Block', 'Cost', 'Memory', 'CPU', 'Latency'], numalign='decimal', stralign='center'))


def print_cpath_stat(tree: nx.DiGraph, partition: T_PART, cpath: list[int] = None, delay: int = 10):
    """
    Print the related block of the critical path.

    :param tree:        input tree
    :param partition:   given partitioning
    :param cpath:       critical path
    :param delay:       platform delay value
    """
    if len(partition) > 0:
        c_blocks = path_blocks(partition, cpath)
        opt_cut = len(c_blocks) - 1
        sum_lat = sum(block_latency([tree.nodes[v][RUNTIME] for v in blk], 0, len(blk) - 1, delay, 0, len(blk) - 1)
                      for blk in c_blocks) + opt_cut * delay
        print("Critical blocks of cpath", [cpath[0], cpath[-1]], "=>", c_blocks, "-", "opt_cut:", opt_cut, "-",
              "opt_lat:", sum_lat)


def evaluate_tree_partitioning(tree: nx.DiGraph, partition: T_PART, opt_cost: int, root: int, cp_end: int, M: int,
                               N: int, L: int, delay: int, unit: int):
    """
    Evaluate tree partitioning and print its characteristics.

    :param tree:        input tree
    :param partition:   given partitioning
    :param opt_cost:    optimal partitioning cost
    :param root:        root node
    :param cp_end:      end node of critical path
    :param M:           upper memory limit
    :param N:           CPU count
    :param L:           latency limit
    :param delay:       platform invocation delay
    :param unit:        rounding unit
    """
    tree = leaf_label_nodes(tree)
    print(tree.graph.get(NAME, "tree").center(80, '#'))
    print("Runtime:", [tree.nodes[v][RUNTIME] for v in tree.nodes if v is not PLATFORM])
    print("Memory:", [tree.nodes[v][MEMORY] for v in tree.nodes if v is not PLATFORM])
    print("Rate:", [tree[next(tree.predecessors(v))][v][RATE] for v in tree.nodes if v is not PLATFORM])
    print(f"Tree partitioning [M={M}, N={N}, L={L}:{(root, cp_end)}] => {partition} - opt_cost: {opt_cost}")
    if partition:
        print_cpath_stat(tree, partition, list(ichain(tree, root, cp_end)), delay)
        print_tree_block_stat(tree, partition, unit)
        draw_tree(tree, partition, draw_blocks=True, draw_weights=False)
    print('#' * 80)


########################################################################################################################


def print_ser_tree_block_stat(tree: nx.DiGraph, partition: T_PART, cpath: list[int]):
    """
    Print cost memory and latency values of partition blocks in tabulated format.

    :param tree:        input tree
    :param partition:   given partitioning
    :param cpath:       critical path
    """
    stat = []
    for blk in partition:
        blk_cost = ser_subtree_cost(tree, blk[0], blk)
        blk_lat = ser_subchain_latency(tree, blk[0], set(blk), set(cpath))
        stat.append([str([blk[0], blk[-1]]), blk_cost, sum(tree.nodes[v][MEMORY] for v in blk), blk_lat])
    print(tabulate.tabulate(stat, ['Block', 'Cost', 'Memory', 'Latency'], numalign='decimal', stralign='center'))


def print_ser_cpath_stat(tree: nx.DiGraph, partition: T_PART, cpath: list[int] = None, delay: int = 10):
    """
    Print the related block of the critical path.

    :param tree:        input tree
    :param partition:   given partitioning
    :param cpath:       critical path
    :param delay:       platform delay value
    """
    cpath = set(cpath)
    if len(partition) > 0:
        restricted_blk = [blk for blk in partition if blk[0] in cpath]
        blk_lats = [ser_subchain_latency(tree, blk[0], set(blk), cpath) for blk in restricted_blk]
        sum_lat = sum(blk_lats) + (len(restricted_blk) - 1) * delay
        print("Critical blocks wrt. cpath:", sorted(cpath), "=>", restricted_blk, "-", "opt_lat:", sum_lat)


def evaluate_ser_tree_partitioning(tree: nx.DiGraph, partition: T_PART, opt_cost: int, opt_lat: int, root: int,
                                   cp_end: int, M: int, L: int, delay: int, draw: bool = True):
    """
    Evaluate tree partitioning and print its characteristics assuming serialized platform execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param opt_cost:    optimal partitioning cost
    :param opt_lat:     latency value of the partitioning
    :param root:        root node
    :param cp_end:      end node of critical path
    :param M:           upper memory limit
    :param L:           latency limit
    :param delay:       platform invocation delay
    :param draw:        draw tree
    """
    tree = leaf_label_nodes(tree)
    print(tree.graph.get(NAME, "tree").center(80, '#'))
    print("Runtime:", [tree.nodes[v][RUNTIME] for v in tree.nodes if v is not PLATFORM])
    print("Memory:", [tree.nodes[v][MEMORY] for v in tree.nodes if v is not PLATFORM])
    print("Data:", [tree[next(tree.predecessors(v))][v][DATA] for v in tree.nodes if v is not PLATFORM])
    print("Rate:", [tree[next(tree.predecessors(v))][v][RATE] for v in tree.nodes if v is not PLATFORM])
    print(f"Tree partitioning [M={M}, L={L}:{(root, cp_end)}] => {partition} - opt_cost: {opt_cost},"
          f" opt_lat: {opt_lat}")
    if partition:
        print_ser_cpath_stat(tree, partition, list(ichain(tree, root, cp_end)), delay)
        print(f"Recalculated partition cost: {sum(ser_subtree_cost(tree, blk[0], blk) for blk in partition)}")
        print_ser_tree_block_stat(tree, partition, set(ichain(tree, root, cp_end)))
        if draw:
            draw_tree(tree, partition, draw_blocks=True, draw_weights=False)
    print('#' * 80)


def print_par_tree_block_stat(tree: nx.DiGraph, partition: T_PART, cpath: list[int], N: int = 1):
    """
    Print cost memory and latency values of partition blocks in tabulated format  assuming parallelized execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param cpath:       critical path
    :param N:           CPU count
    """
    stat = []
    for blk in partition:
        blk_cost = par_subtree_cost(tree, blk[0], blk, N)
        blk_lat = par_subchain_latency(tree, blk[0], set(blk), set(cpath), N)
        stat.append([str([blk[0], blk[-1]]), blk_cost, sum(tree.nodes[v][MEMORY] for v in blk), blk_lat])
    print(tabulate.tabulate(stat, ['Block', 'Cost', 'Memory', 'Latency'], numalign='decimal', stralign='center'))


def print_par_cpath_stat(tree: nx.DiGraph, partition: T_PART, cpath: list[int] = None, delay: int = 10, N: int = 1):
    """
    Print the related block of the critical path assuming parallelized execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param cpath:       critical path
    :param delay:       platform invocation delay
    :param N:           CPU count
    """
    cpath = set(cpath)
    if len(partition) > 0:
        restricted_blk = [blk for blk in partition if blk[0] in cpath]
        blk_lats = [par_subchain_latency(tree, blk[0], set(blk), cpath, N) for blk in restricted_blk]
        sum_lat = sum(blk_lats) + (len(restricted_blk) - 1) * delay
        print("Critical blocks wrt. cpath:", sorted(cpath), "=>", restricted_blk, "-", "opt_lat:", sum_lat)


def evaluate_par_tree_partitioning(tree: nx.DiGraph, partition: T_PART, opt_cost: int, opt_lat: int, root: int,
                                   cp_end: int, M: int, L: int, N: int, delay: int, draw: bool = True):
    """
    Evaluate tree partitioning and print its characteristics assuming parallelized platform execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param opt_cost:    optimal partitioning cost
    :param opt_lat:     latency value of the partitioning
    :param root:        root node
    :param cp_end:      end node of critical path
    :param M:           upper memory limit
    :param L:           latency limit
    :param N:           CPU count
    :param delay:       platform invocation delay
    :param draw:        draw tree
    """
    tree = leaf_label_nodes(tree)
    print(tree.graph.get(NAME, "tree").center(80, '#'))
    print("Runtime:", [tree.nodes[v][RUNTIME] for v in tree.nodes if v is not PLATFORM])
    print("Memory:", [tree.nodes[v][MEMORY] for v in tree.nodes if v is not PLATFORM])
    print("Data:", [tree[next(tree.predecessors(v))][v][DATA] for v in tree.nodes if v is not PLATFORM])
    print("Rate:", [tree[next(tree.predecessors(v))][v][RATE] for v in tree.nodes if v is not PLATFORM])
    print(f"Tree partitioning [{M=}, {L=}:{(root, cp_end)}, {N=}] => {partition} - opt_cost: {opt_cost},"
          f" opt_lat: {opt_lat}")
    if partition:
        print_par_cpath_stat(tree, partition, list(ichain(tree, root, cp_end)), delay, N)
        print_par_tree_block_stat(tree, partition, set(ichain(tree, root, cp_end)), N)
        if draw:
            draw_tree(tree, partition, draw_blocks=True, draw_weights=False)
    print('#' * 80)


def evaluate_gen_tree_partitioning(tree: nx.DiGraph, partition: T_PART, opt_cost: int, opt_lat: int, root: int,
                                   flavors: list, cp_end: int, L: int, delay: int, draw: bool = True):
    """
    Evaluate tree partitioning and print its characteristics assuming parallelized platform execution model.

    :param tree:        input tree
    :param partition:   given partitioning
    :param opt_cost:    optimal partitioning cost
    :param opt_lat:     latency value of the partitioning
    :param root:        root node
    :param flavors:     list of flavors
    :param cp_end:      end node of critical path
    :param L:           latency limit
    :param delay:       platform invocation delay
    :param draw:        draw tree
    """
    tree = leaf_label_nodes(tree)
    print(tree.graph.get(NAME, "tree").center(80, '#'))
    print("Runtime:", [tree.nodes[v][RUNTIME] for v in tree.nodes if v is not PLATFORM])
    print("Memory:", [tree.nodes[v][MEMORY] for v in tree.nodes if v is not PLATFORM])
    print("Data:", [tree[next(tree.predecessors(v))][v][DATA] for v in tree.nodes if v is not PLATFORM])
    print("Rate:", [tree[next(tree.predecessors(v))][v][RATE] for v in tree.nodes if v is not PLATFORM])
    M, N, cfactor = zip(*flavors)
    print(f"Tree partitioning [{M=}, {L=}:{(root, cp_end)}, {N=}] => {partition} - opt_cost: {opt_cost},"
          f" opt_lat: {opt_lat}")
    if partition:
        # print_par_cpath_stat(tree, partition, list(ichain(tree, root, cp_end)), delay, N)
        # print_par_tree_block_stat(tree, partition, set(ichain(tree, root, cp_end)), N)
        if draw:
            draw_tree(tree, partition, draw_blocks=True, draw_weights=False)
    print('#' * 80)


def print_ser_chain_summary(runtime: list[int], memory: list[int], rate: list[int], data: list[int]):
    """
    Print chain summary assuming serialized execution model.

    :param runtime:     list of runtime values
    :param memory:      list of memory values
    :param rate:        list of rate values
    :param data:        list of data values
    """
    print("Chain:", "[", *(f"-{r}-> F(D{d}|T{t}|M{m})" for t, m, r, d in zip(runtime, memory, rate, data)), "]")


def print_ser_block_stat(partition: T_PART, runtime: list[int], memory: list[int], rate: list[int], data: list[int],
                         delay: float, start: int = 0, end: int = None):
    """
    Print block stats of a chain partitioning assuming serialized execution model.

    :param partition:   given partitioning
    :param runtime:     list of runtime values
    :param memory:      list of memory values
    :param rate:        list of rate values
    :param data:        list of data values
    :param delay:       platform delay
    :param start:       fist node to consider
    :param end:         last node to consider
    """
    end = end if end is not None else len(runtime) - 1
    stat = [[str([blk[0], blk[-1]]),
             ser_block_subcost(runtime, rate, data, blk[0], blk[-1]),
             ser_block_submemory(memory, blk[0], blk[-1]),
             ser_block_sublatency(runtime, rate, data, blk[0], blk[-1], delay, start, end)] for blk in partition]
    print(tabulate.tabulate(stat, ['Block', 'Cost', 'Memory', 'Latency'],
                            numalign='decimal', stralign='center', tablefmt='pretty'))


def evaluate_ser_chain_partitioning(partition: T_PART, opt_cost: int, opt_lat: int, runtime: list[int],
                                    memory: list[int], rate: list[int], data: list[int], M: int = math.inf,
                                    L: int = math.inf, start: int = 0, end: int = None, delay: int = 1):
    """
    Evaluate chain partitioning and print its characteristics assuming serialized execution model.

    :param partition:   given partitioning
    :param opt_cost:    optimal partitioning cost
    :param opt_lat:     latency value of the partitioning
    :param runtime:     list of runtime values
    :param memory:      list of memory values
    :param rate:        list of rate values
    :param data:        list of data values
    :param M:           upper memory limit
    :param L:           latency limit
    :param start:       fist node to consider
    :param end:         last node to consider
    :param delay:       platform delay
    """
    print('#' * 80)
    print(f"Chain partitioning [M={M}, L={L}:{(start, end)}] => "
          f"{partition} - opt_cost: {opt_cost}, opt_lat: {opt_lat}")
    print_ser_block_stat(partition, runtime, memory, rate, data, delay, start, end)
    print('#' * 80)


########################################################################################################################


def print_lp_desc(model: pulp.LpProblem):
    """
    Print the lp format of the model.

    :param model:   PuLP model object
    """
    with tempfile.TemporaryDirectory() as tmp:
        model.writeLP(f"{tmp}/chain_model.lp")
        with open(f"{tmp}/chain_model.lp") as f:
            print(f.read())


def convert_var_dict(X: dict[int, dict[int]]) -> list[list[pulp.LpVariable]]:
    """
    Convert dict-of-dict variable matrix into list-of-list format.

    :param X:   specific structure of decision variables
    :return:    converted format of decision variables
    """
    return [[X[i][j] if j in X[i] else None for j in range(1, i + 1)] for i in sorted(X)]


def print_var_matrix(X: list[list[pulp.LpVariable]]):
    """
    Print matrix of decision variables names in tabular format.

    :param X:   specific structure of decision variables
    """
    print(tabulate.tabulate([list(map(lambda x: x if isinstance(x, (int, type(None))) else x.name, x_i)) for x_i in X],
                            missingval="-", numalign='center', stralign='center', tablefmt='outline'))


def print_pulp_matrix_values(X: list[list[pulp.LpVariable]]):
    """
    Print matrix of decision variables values in tabular format.

    :param X:   specific structure of decision variables
    """
    print(tabulate.tabulate([list(map(lambda x: x if isinstance(x, (int, type(None))) else pulp.value(x), x_i))
                             for x_i in X], missingval="-", numalign='center', stralign='center', tablefmt='outline'))


def print_cplex_matrix_values(X: list[list[pulp.LpVariable]]):
    """
    Print matrix of decision variables values in tabular format.

    :param X:   specific structure of decision variables
    """
    print(tabulate.tabulate([list(map(lambda x: x if isinstance(x, (int, type(None))) else round(x.solution_value),
                                      x_i)) for x_i in X],
                            missingval="-", numalign='center', stralign='center', tablefmt='outline'))


def print_cost_coeffs(model: pulp.LpProblem, X: list[list[pulp.LpVariable]]):
    """
    Print cost coefficients of the given LP *model*.

    :param model:   model object
    :param X:       specific structure of decision variables
    """
    print(tabulate.tabulate([[model.objective.get(x) for x in x_i] for x_i in X],
                            missingval="-", numalign='center', stralign='center', tablefmt='outline'))


def print_lat_coeffs(model: pulp.LpProblem, X: list[list[pulp.LpVariable]]):
    """
    Print latency coefficients of the given LP *model*.

    :param model:   model object
    :param X:       specific structure of decision variables
    """
    print(tabulate.tabulate([[model.constraints[LP_LAT].get(x) for x in x_i] for x_i in X],
                            missingval="-", numalign='center', stralign='center', tablefmt='outline'))



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import pprint
import sys
import time

import networkx as nx
import numpy as np
import pulp
import pytest
import tabulate

from slambuc.alg import LP_LAT
from slambuc.alg.service import NAME, PLATFORM
from slambuc.alg.tree.ser.ilp import (build_tree_mtx_model, tree_mtx_partitioning, recreate_subtrees_from_xmatrix,
                                      extract_subtrees_from_xmatrix, build_greedy_tree_mtx_model)
from slambuc.alg.util import induced_subtrees, ibacktrack_chain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.plot import draw_tree
from slambuc.misc.util import (print_lp_desc, evaluate_ser_tree_partitioning, print_var_matrix, get_cplex_path,
                               print_pulp_matrix_values, convert_var_dict, print_cost_coeffs, print_lat_coeffs,
                               get_glpk_path)


def test_reachable_nodes(branch: int = 2, depth: int = 3, random_nodes: int = 0):
    if random_nodes:
        tree = get_random_tree(nodes=random_nodes)
    else:
        tree = nx.balanced_tree(branch, depth, create_using=nx.DiGraph)
        tree.graph[NAME] = f"balanced_tree({branch=},{depth=})"
        tree.add_edge(PLATFORM, 0)
    print("\nReachable nodes:")
    for (_, n), nodes in induced_subtrees(tree, int(bool(random_nodes)), only_nodes=True):
        print(n, '-->', nodes)
    print("\nReachable edges:")
    for (_, n), edges in induced_subtrees(tree, int(bool(random_nodes)), only_nodes=False):
        print(n, '-->', edges)
    reachable = {n: r for (_, n), r in induced_subtrees(tree, int(bool(random_nodes)), only_nodes=True)}
    rm = np.full((len(tree) - 1, len(tree) - 1), None)
    for i, r in reachable.items():
        for j in r:
            coord = (i - 1, j - 1) if random_nodes else (i, j)
            rm[coord] = 1
    print("Reachability Matrix (R):")
    print(tabulate.tabulate(rm, tablefmt='outline'))
    print("Connectivity coefficients (R^T):")
    print(tabulate.tabulate(rm.transpose(), tablefmt='outline'))
    draw_tree(tree)


def test_mtx_model_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml",
                            save_file: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  subchains=True,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    model_greedy, X_greedy = build_greedy_tree_mtx_model(**params)
    _d = time.perf_counter() - _s
    print(f"Greedy Model building time: {_d * 1000} ms")
    _s = time.perf_counter()
    model, X = build_tree_mtx_model(**params)
    _d = time.perf_counter() - _s
    print(f"Direct Model building time: {_d * 1000} ms")
    X = convert_var_dict(X)
    X_greedy = convert_var_dict(X_greedy)
    print("  Decision variables  ".center(80, '='))
    print("Greedy:")
    print_var_matrix(X_greedy)
    print("Direct:")
    print_var_matrix(X)
    print("  Cost Coefficients  ".center(80, '='))
    print("Greedy:")
    print_cost_coeffs(model_greedy, X_greedy)
    print("Direct:")
    print_cost_coeffs(model, X)
    print("  Latency Coefficients  ".center(80, '='))
    print("Greedy:")
    print_lat_coeffs(model_greedy, X_greedy)
    print("Direct:")
    print_lat_coeffs(model, X)
    print("  Generated LP model  ".center(80, '='))
    print_lp_desc(model_greedy)
    print_lp_desc(model)
    if save_file:
        model.writeLP("tree_mtx_model.lp")


def test_mtx_model_solution(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml"):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    #
    print(" Greedy model ".center(80, '='))
    model, X = build_greedy_tree_mtx_model(**params)
    status = model.solve(solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True))
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    print("Solution:")
    print_pulp_matrix_values(convert_var_dict(X))
    #
    print(" Direct model ".center(80, '='))
    model, X = build_tree_mtx_model(**params)
    solver = pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True)
    status = model.solve(solver=solver)
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    #
    print("Solution:")
    print_pulp_matrix_values(convert_var_dict(X))
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    _s = time.perf_counter()
    rec_partition = recreate_subtrees_from_xmatrix(tree, X)
    _d = time.perf_counter() - _s
    print(f"Recreate: {_d * 1000} ms")
    _s = time.perf_counter()
    ext_partition = extract_subtrees_from_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {rec_partition = }, {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
@pytest.mark.skipif(not (sys.version_info < (3, 11)), reason="PY version is not supported!")
def test_mtx_model_solution_cplex():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  delay=10)
    print("  Run CPLEX solver  ".center(80, '='))
    model, X = build_tree_mtx_model(**params)
    solver = pulp.CPLEX_PY(mip=True, warmStart=False, msg=True)
    status = model.solve(solver=solver)
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    if status == pulp.LpStatusOptimal:
        opt_cost, opt_lat = pulp.value(model.objective), pulp.value(model.constraints[LP_LAT])
        partition = extract_subtrees_from_xmatrix(X)
        opt_lat = params['L'] + opt_lat if params['L'] < math.inf else opt_lat
        print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
        print(f"{model.solutionTime    = } s")
        print(f"{model.solutionCpuTime = } s")
        print(f"{solver.solveTime = } s")
        #
        print("*" * 30)
        print(f"CPLEX model: {solver.solverModel}")
        print("CPLEX stat:", solver.solverModel.get_stats())


@pytest.mark.skipif(get_glpk_path() is None, reason="GLPK is not available!")
def test_mtx_model_solution_glpk():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  delay=10)
    print("  Run GLPK solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_mtx_partitioning(**params, solver=pulp.GLPK_CMD(mip=True, msg=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")


def evaluate_ilp_mtx_model():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  delay=10)
    print("  CBC solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_mtx_partitioning(**params, solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)
    print("  CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_mtx_partitioning(**params, solver=pulp.CPLEX_PY(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)


def evaluate_ilp_mtx_subchains_model():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  delay=10)
    print("  CBC solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_mtx_partitioning(**params, subchains=True,
                                                         solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)
    print("  CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_mtx_partitioning(**params, subchains=True,
                                                         solver=pulp.CPLEX_PY(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)


########################################################################################################################


def run_test(tree: nx.DiGraph, root: int, cp_end: int, M: int, L: int, delay: int):
    partition, opt_cost, opt_lat = tree_mtx_partitioning(tree, root, M, L, cp_end, delay,
                                                         solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=False))
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    return partition, opt_cost, opt_lat


def test_ser_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  # L = math.inf
                  L=430,
                  delay=10)
    run_test(**params)


def test_random_ser_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_reachable_nodes(random_nodes=0)
    test_mtx_model_creation()
    # test_mtx_model_creation(tree_file="failed_tree_random_tree_1673972740.2507007.gml")
    # test_mtx_model_solution()
    # test_mtx_model_solution_cplex()
    # test_mtx_model_solution_glpk()
    # evaluate_ilp_mtx_model()
    # evaluate_ilp_mtx_subchains_model()
    # test_ser_tree()
    # test_random_ser_tree()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import functools
import math
import typing

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS, T_PART
from slambuc.alg.service import *
from slambuc.alg.util import ipostorder_dfs, ibacktrack_chain, verify_limits


class TBlock(typing.NamedTuple):
    """Store subtree attributes for a given seq subcase."""
    w: int = None  # Tailing node of the first block of the subtree partitioning
    sum_cost: int = math.inf  # Sum cost of the subtree partitioning
    cumsum: int = 0  # Sum (cumulative) runtime of the first block (with tail node w) in the partitioning
    mem: int = math.inf  # Sum memory of the first block
    max_rate: int = 0  # Maximum rate value of internal edge in the first block
    cpu: int = 1  # Sum CPU core need of the first block

    def __repr__(self):
        return repr(tuple(self))


def seq_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = math.inf,
                          L: int = math.inf, cp_end: int = None, delay: int = 1, unit: int = 100,
                          full: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes leveraging
    a bottom-up tree traversal approach.

    Cost calculation relies on the rounding *unit* and number of vCPU cores *N*, whereas platform invocation *delay*
    is used for latency calculations.

    Details in: J. Czentye, I. Pelle and B. Sonkoly, "Cost-optimal Operation of Latency Constrained Serverless
    Applications: From Theory to Practice," NOMS 2023-2023 IEEE/IFIP Network Operations and Management Symposium,
    Miami, FL, USA, 2023, pp. 1-10, doi: 10.1109/NOMS56928.2023.10154412.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :param full:    return full blocks or just their ending nodes
    :return:        tuple of optimal partition, sum cost of the partitioning, and optimal number of cuts
    """
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # c_max is the number of cuts allowed by L or at most the number of edges on cpath
    c_max = math.floor(min((L - sum(tree.nodes[_v][RUNTIME] for _v in cpath)) / delay, len(cpath) - 1))
    # Check lower bound for latency limit
    if c_max < 0:
        return [], None, c_max
    DP = [collections.defaultdict(collections.deque) for _ in range(len(tree))]
    for i in range(len(DP)):
        DP[i][0].append(TBlock())

    @functools.lru_cache(maxsize=(len(tree) - 1))
    def block_cost(pred: int, barr: int, cumsum: int, expand: bool = True) -> tuple[int, int]:
        """Calculate sum cost of subtree: T_barr and also return the cumulative sum runtime of the block[barr, w]."""
        if expand:
            cumsum += tree.nodes[barr][RUNTIME]
        return tree[pred][barr][RATE] * (math.ceil(cumsum / unit) * unit), cumsum

    @functools.lru_cache(maxsize=(len(tree) - 1))
    def block_cpu(pred: int, node: int, max_rate: int, cpu: int) -> tuple[int, int]:
        """Calculate the nex CPU core need of the block[barr, w] and also return the max internal rate."""
        blk_max_rate = max(max_rate, tree[pred][node][RATE])
        return max(cpu, math.ceil(blk_max_rate / tree[pred][node][RATE])), blk_max_rate

    def qmin(node: int, c_n: int) -> int:
        """Return the sum cost of best/min subcase for *node* with *c_n* cuts."""
        return DP[node][c_n][0].sum_cost

    def qinsert(node: int, c_n: int, blk: TBlock):
        """Insert given block subcase *block* into *DP* for the *node* with *c_n* cuts."""
        if blk.sum_cost < math.inf and blk.mem <= M and blk.cpu <= N:
            if len(DP[node][c_n]) and blk.sum_cost <= qmin(node, c_n):
                DP[node][c_n].appendleft(blk)
            else:
                DP[node][c_n].append(blk)

    def qmerge(pred: int, node: int, c_n: int, barr: int, c_b: int, m_cost: int):
        """
        Copy DP entries from queue of node *barr* with *c_b* cuts into queue of node *node* with *c_n* cuts
        while leaving the best subcase in the original queue.
        """
        for blk in DP[barr][c_b]:
            # Ignore infeasible subcases
            if blk.sum_cost < math.inf:
                # Calculate the original cost of the block[barr, w]
                b_blk_cost, _ = block_cost(node, barr, blk.cumsum, expand=False)
                # Calculate the cost of the expanded block[node, w], n -> barr
                n_blk_cost, n_blk_cumsum = block_cost(pred, node, blk.cumsum)
                # Calculate the new sum_cost
                n_sum_cost = blk.sum_cost + (n_blk_cost - b_blk_cost) + m_cost
                # Calculate the new memory
                n_blk_mem = blk.mem + tree.nodes[node][MEMORY]
                # Calculate the new CPU need
                blk_cpu, blk_max_rate = block_cpu(pred, node, blk.max_rate, blk.cpu)
                qinsert(node, c_n, TBlock(blk.w, n_sum_cost, n_blk_cumsum, n_blk_mem, blk_max_rate, blk_cpu))
        # If no feasible solution exists with c cuts -> add default with infinity cost
        if not DP[node][c_n]:
            DP[node][c_n].append(TBlock())

    def qclear(node: int, c_n: int):
        """Leave only the best/min subcase in the queue as the first element"""
        best_blk = DP[node][c_n][0]
        DP[node][c_n].clear()
        DP[node][c_n].append(best_blk)

    for p, n in ipostorder_dfs(tree, root):
        n_mem, n_rate = tree.nodes[n][MEMORY], tree[p][n][RATE]
        # Subcases of leaves can be precalculated to store the single block -> [n]
        if len(tree.succ[n]) < 1:
            sum_n_cost, n_cumsum = block_cost(p, n, 0)
            qinsert(n, 0, TBlock(n, sum_n_cost, n_cumsum, n_mem, n_rate, 1))
            continue
        # Sum best subcases of n's successors not involved in cpath
        sum_m_cost = sum(qmin(m, 0) for m in tree.succ[n] if m not in cpath)
        if n not in cpath:
            # Single block subcase -> [n] + sum(m): n -> m
            n_cost, n_cumsum = block_cost(p, n, 0)
            sum_n_cost = n_cost + sum_m_cost
            qinsert(n, 0, TBlock(n, sum_n_cost, n_cumsum, n_mem, n_rate, 1))
            # Merged subcases -> [n] U [b -> w] + sum(m): n -> b, n -> m, m != b
            for b in tree.succ[n]:
                qmerge(p, n, 0, b, 0, m_cost=sum_m_cost - qmin(b, 0))
                qclear(b, 0)
        else:
            # Cut subcase -> [n] + m_cp + sum(m\m_cp): n -> m, m != m_cp
            n_cost, n_cumsum = block_cost(p, n, 0)
            m_cp = next(m for m in tree.succ[n] if m in cpath)
            # Since n -> b is a cut, at most c_max-1 subcases should be referenced
            for c in range(1, min(len(DP[m_cp]), c_max) + 1):
                sum_n_cost = n_cost + sum_m_cost + qmin(m_cp, c - 1)
                qinsert(n, c, TBlock(n, sum_n_cost, n_cumsum, n_mem, n_rate, 1))
            # Merged subcases -> [n] U [b -> w] + sum(m\m_cp) + m_cp: n -> b, n -> m, m != b != m_cp
            for b in tree.succ[n]:
                if b == m_cp:
                    for c in range(0, min(len(DP[m_cp]), c_max + 1)):
                        qmerge(p, n, c, b, c, m_cost=sum_m_cost)
                        qclear(b, c)
                else:
                    m_res_cost = sum_m_cost - qmin(b, 0)
                    # Subcases of node b with 0 cut is reused to calculate subcases with different number of cuts
                    for c in range(1, min(len(DP[m_cp]), c_max) + 1):
                        qmerge(p, n, c, b, 0, m_cost=m_res_cost + qmin(m_cp, c - 1))
                    qclear(b, 0)
    c_opt = min(DP[root], key=lambda _c: qmin(root, _c))
    if qmin(root, c_opt) < math.inf:
        return extract_blocks(tree, DP, root, cp_end, c_opt, full), qmin(root, c_opt), c_opt
    else:
        return [], math.inf, c_opt


def extract_blocks(tree: nx.DiGraph, DP: list[collections.defaultdict[collections.deque[TBlock]]],
                   root: int, cp_end: int, c_opt: int, full: bool = True) -> T_PART:
    """
    Extract subtree roots of partitioning from the tailing nodes stored in the *DP* matrix.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param DP:      dynamic programming structure storing intermediate *TBlock* subcases
    :param root:    root node of the graph
    :param cp_end:  tail node of the critical path
    :param c_opt:   calculated optimal cut size
    :param full:    calculate full blocks
    :return:        partitioning blocks
    """
    n = set(filter(lambda v: v is not PLATFORM, tree))
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    p = []
    barr = {(root, c_opt)}
    while len(n):
        b, c = barr.pop()
        w = DP[b][c][0].w
        blk, prior = [], None
        while prior != b:
            for m in tree.succ[w]:
                if m != prior:
                    barr.add((m, c - 1) if m in cpath else (m, 0))
            if full:
                blk.append(w)
            n.remove(w)
            prior = w
            w = next(tree.predecessors(w))
        if blk[-1] != b:
            blk.append(b)
        blk.reverse()
        p.append(blk)
    return sorted(p)



# Copyright 2024 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import random

from slambuc.alg.chain.sp.min import sp_chain_partitioning
from slambuc.alg.util import split_chain
from slambuc.misc.util import evaluate_chain_partitioning, print_chain_summary


def run_test(runtime: list, memory: list, rate: list, M: int = math.inf, N: int = math.inf,
             L: int = math.inf, delay: int = 1, unit: int = 100):
    barr, opt_cost, opt_lat = sp_chain_partitioning(runtime, memory, rate, M, N, L, delay, unit)
    partition = split_chain(barr, len(runtime)) if barr else []
    evaluate_chain_partitioning(partition, opt_cost, opt_lat, runtime, memory, rate, M, N, L, 0, len(runtime),
                                delay, unit)
    return partition, opt_cost, opt_lat


def test_chain():
    runtime = [20, 40, 50, 20, 70, 40, 50, 60, 40, 10]
    memory = [3, 3, 2, 1, 2, 1, 2, 1, 2, 3]
    rate = [1, 1, 2, 2, 1, 3, 1, 2, 1, 3]
    delay = 10
    M = 6
    N = 3
    L = 450
    unit = 100
    print_chain_summary(runtime, memory, rate)
    run_test(**locals())


def test_random_chain():
    runtime = [random.randint(10, 100) for _ in range(10)]
    memory = [random.randint(1, 3) for _ in range(10)]
    rate = [random.randint(1, 3) for _ in range(10)]
    delay = 10
    M = 6
    N = 3
    L = sum(runtime) + 1 + delay * random.randint(len(runtime) // 4, len(runtime) // 2)
    unit = 100
    print_chain_summary(runtime, memory, rate)
    run_test(**locals())


if __name__ == '__main__':
    test_chain()
    # test_random_chain()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib

import networkx as nx
import pytest

from slambuc.alg.service import NAME
from slambuc.alg.tree.dp.greedy import greedy_tree_partitioning
from slambuc.alg.util import leaf_label_nodes
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import evaluate_tree_partitioning


def run_test(tree: nx.DiGraph, M: int, N: int, L: int, root: int = 1, cp_end: int = None, delay: int = 10,
             unit: int = 100):
    leaf_label_nodes(tree)
    results = greedy_tree_partitioning(tree, root, M, N, L, cp_end, delay, unit)
    for i, (part, best_cost, best_lat) in enumerate(results):
        print(f"  GREEDY[{i}]  ".center(80, '#'))
        evaluate_tree_partitioning(tree, part, best_cost, root, cp_end, M, N, L, delay, unit)
    return results


def test_greedy_partitioning():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    tree.graph[NAME] += "-greedy"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=15,
                  N=2,
                  L=math.inf,
                  delay=10)
    run_test(**params)


@pytest.mark.skip("No failed input tree is provided.")
def test_failed_greedy_partitioning(graph_path: str, L=math.inf):
    tree = nx.read_gml(graph_path, destringizer=int)
    tree.graph[NAME] += "-greedy"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  N=2,
                  L=L,
                  delay=10)
    run_test(**params)


def test_random_greedy_partitioning(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-greedy"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  N=2,
                  L=math.inf,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    test_greedy_partitioning()
    # test_failed_greedy_partitioning("failed_tree_random_tree_1659449915.340366.gml", L=396)
    # test_random_greedy_partitioning()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import math
import typing

import networkx as nx

from slambuc.alg import INFEASIBLE, T_BLOCK, T_RESULTS
from slambuc.alg.service import *
from slambuc.alg.util import ipostorder_dfs, ibacktrack_chain


class MinTBlock(typing.NamedTuple):
    """Store subtree attributes for a given min subcase."""
    w: int = None  # Tailing node of the first block of the subtree partitioning
    c: int = 0  # Number of cuts the given subtree partitioning introduce on the critical path
    sum_cost: int = math.inf  # Sum cost of the subtree partitioning
    cost: int = math.inf  # Cost of the first block (with tail node w) in the subtree partitioning
    mem: int = math.inf  # Sum memory of the first block
    max_rate: int = 0  # Maximum rate value of internal edge in the first block
    cpu: int = 1  # Sum CPU core need of the first block

    def __repr__(self):
        return repr(tuple(self))


def min_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = math.inf, L: int = math.inf,
                          cp_end: int = None, delay: int = 1, unit: int = 100, full: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes.

    Cost calculation relies on the rounding *unit* and number of vCPU cores *N*, whereas platform invocation *delay*
    is used for latency calculations.

    It gives optimal result only in case the cost function regarding the chain attributes is sub-additive,
    that is k_opt = k_min is guaranteed for each case.

    Instead of full partitioning it only returns the list of barrier nodes.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :param full:    return full blocks or just their ending nodes
    :return:        tuple of optimal partition, sum cost of the partitioning, and optimal number of cuts
    """
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # c_max is the number of cuts allowed by L or at most the number of edges on cpath
    c_max = math.floor(min((L - sum(tree.nodes[_v][RUNTIME] for _v in cpath)) / delay, len(cpath) - 1))
    # Check lower bound for latency limit
    if c_max < 0:
        return [], None, c_max
    DP = [collections.deque((MinTBlock(),)) for _ in range(len(tree))]

    def qmin(node: int) -> MinTBlock:
        """Return the sum cost of best/min subcase for *node* with *c_n* cuts."""
        return DP[node][0]

    def qinsert(node: int, blk: MinTBlock):
        """Insert given block subcase *block* into *DP* for the *node* with *c_n* cuts."""
        if blk.sum_cost < math.inf and blk.mem <= M and blk.cpu <= N and blk.c <= c_max:
            if blk.sum_cost <= qmin(node).sum_cost:
                DP[node].appendleft(blk)
            else:
                DP[node].append(blk)

    def qmerge(pred: int, node: int, barr: int, m_cost: int, add_cut: bool = False):
        """
        Copy DP entries from queue of node *barr* with *c_b* cuts into queue of node *node* with *c_n* cuts
        while leaving the best subcase in the original queue.
        """
        blk = None
        while len(DP[barr]):
            blk = DP[barr].pop()
            # Ignore infeasible subcases
            if blk.sum_cost < math.inf:
                blk_cost = block_cost(pred, node, blk.w)
                sum_blk_cost = blk.sum_cost + (blk_cost - blk.cost) + m_cost
                sum_blk_cut = blk.c + 1 if add_cut else blk.c
                blk_mem = blk.mem + tree.nodes[node][MEMORY]
                blk_max_rate = max(blk.max_rate, tree[pred][node][RATE])
                blk_cpu = max(blk.cpu, math.ceil(blk_max_rate / tree[pred][node][RATE]))
                qinsert(node, MinTBlock(blk.w, sum_blk_cut, sum_blk_cost, blk_cost, blk_mem, blk_max_rate, blk_cpu))
        # Leave only the best/min subcase in the queue as the first element
        DP[barr].append(blk)

    def block_cost(pred: int, barr: int, w: int) -> int:
        """Calculate running time of block: p -> [barr -> w]."""
        sum_time = tree.nodes[w][RUNTIME]
        while w > barr:
            w = next(tree.predecessors(w))
            sum_time += tree.nodes[w][RUNTIME]
        return tree[pred][barr][RATE] * (math.ceil(sum_time / unit) * unit)

    for p, n in ipostorder_dfs(tree, root):
        n_mem, n_rate = tree.nodes[n][MEMORY], tree[p][n][RATE]
        # Subcases of leaves can be precalculated to store the single block -> [n]
        if len(tree.succ[n]) < 1:
            n_cost = block_cost(p, n, n)
            qinsert(n, MinTBlock(n, 0, n_cost, n_cost, n_mem, n_rate, 1))
            continue
        # Sum best subcases of n's successors not involved in cpath
        sum_m_cost = sum(qmin(m).sum_cost for m in tree.succ[n])
        if n not in cpath:
            # Single block subcase -> [n] + sum(m): n -> m
            n_cost = block_cost(p, n, n)
            qinsert(n, MinTBlock(n, 0, n_cost + sum_m_cost, n_cost, n_mem, n_rate, 1))
            # Merged subcases -> [n] U [b -> w] + sum(m): n -> b, n -> m, m != b
            for b in tree.succ[n]:
                qmerge(p, n, b, sum_m_cost - qmin(b).sum_cost, add_cut=False)
        else:
            # Cut subcase -> [n] + m_cp + sum(m\m_cp): n -> m, m != m_cp
            n_cost = block_cost(p, n, n)
            m_cp = next(m for m in tree.succ[n] if m in cpath)
            # Since n -> b is a cut, at most c_max-1 subcases should be referenced
            qinsert(n, MinTBlock(n, qmin(m_cp).c + 1, n_cost + sum_m_cost, n_cost, n_mem, n_rate, 1))
            # Merged subcases -> [n] U [b -> w] + sum(m\m_cp) + m_cp: n -> b, n -> m, m != b != m_cp
            for b in tree.succ[n]:
                qmerge(p, n, b, sum_m_cost - qmin(b).sum_cost, add_cut=True if b != m_cp else False)
    if qmin(root).sum_cost < math.inf:
        return extract_min_blocks(tree, DP, root, full), qmin(root).sum_cost, qmin(root).c
    else:
        return INFEASIBLE


def extract_min_blocks(tree: nx.DiGraph, DP: list[collections.deque[MinTBlock]],
                       root: int, full: bool = True) -> T_BLOCK:
    """
    Extract subtree roots of partitioning from the tailing nodes stored in the *DP* matrix.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param DP:      dynamic programming structure storing intermediate *MinTBlock* subcases
    :param root:    root node of the graph
    :param full:    calculate full blocks
    :return:        partitioning blocks
    """
    n = set(filter(lambda v: v is not PLATFORM, tree))
    p = []
    barr = {root}
    while len(n):
        b = barr.pop()
        w = DP[b][0].w
        blk, prior = [], None
        while prior != b:
            for m in tree.succ[w]:
                if m != prior:
                    barr.add(m)
            if full:
                blk.append(w)
            n.remove(w)
            prior = w
            w = next(tree.predecessors(w))
        if blk[-1] != b:
            blk.append(b)
        blk.reverse()
        p.append(blk)
    return sorted(p)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
import time

import networkx as nx

from slambuc.alg.service.common import *


def get_random_chain_data(nodes: int = 10, runtime: tuple[int, int] = (1, 100), memory: tuple[int, int] = (1, 3),
                          rate: tuple[int, int] = (1, 3), data: tuple[int, int] = (1, 20)) -> list[list[int]]:
    """
    Generate random chain(path graph) data with properties from given intervals.

    :param nodes:   number of nodes
    :param runtime: interval of runtime values
    :param memory:  interval of memory values
    :param rate:    interval of rate values
    :param data:    interval of data values
    :return:        generated chain data
    """
    t = [random.randint(*runtime) for _ in range(nodes)]
    m = [random.randint(*memory) for _ in range(nodes)]
    d = [random.randint(*data) for _ in range(nodes)]
    r = [1, *(random.randint(*rate) for _ in range(nodes - 1))]
    return t, m, r, d


def get_random_chain(nodes: int = 10, runtime: tuple[int, int] = (1, 100), memory: tuple[int, int] = (1, 3),
                     rate: tuple[int, int] = (1, 3), data: tuple[int, int] = (1, 20)) -> nx.DiGraph:
    """
    Generate random chain(path graph) with properties from given intervals.

    :param nodes:   number of nodes
    :param runtime: interval of runtime values
    :param memory:  interval of memory values
    :param rate:    interval of rate values
    :param data:    interval of data values
    :return:        generated random chain
    """
    chain = nx.path_graph(range(0, nodes + 1), nx.DiGraph)
    nx.set_node_attributes(chain, {i: random.randint(*runtime) for i in range(1, nodes + 1)}, RUNTIME)
    nx.set_node_attributes(chain, {i: random.randint(*memory) for i in range(1, nodes + 1)}, MEMORY)
    for _, _, d in chain.edges(data=True):
        d[RATE] = random.randint(*rate)
        d[DATA] = random.randint(*data)
    chain = nx.relabel_nodes(chain, {0: PLATFORM})
    chain[PLATFORM][1][RATE] = 1
    chain.graph[NAME] = "random_chain"
    return chain


def get_random_tree(nodes: int = 20, runtime: tuple[int, int] = (1, 100), memory: tuple[int, int] = (1, 3),
                    rate: tuple[int, int] = (1, 3), data: tuple[int, int] = (1, 20),
                    name: str = None) -> nx.DiGraph:
    """
    Generate random tree from Prufer sequence with properties from given intervals.

    :param nodes:   number of nodes
    :param runtime: interval of runtime values
    :param memory:  interval of memory values
    :param rate:    interval of rate values
    :param data:    interval of data values
    :param name:    tree name suffix
    :return:        generated random tree
    """
    raw_tree = nx.bfs_tree(nx.random_labeled_tree(nodes + 1), 0)
    while raw_tree.out_degree[0] > 1:
        raw_tree = nx.bfs_tree(nx.random_labeled_tree(nodes + 1), 0)
    tree = nx.convert_node_labels_to_integers(raw_tree, first_label=0)
    nx.set_node_attributes(tree, {i: random.randint(*runtime) for i in range(1, nodes + 1)}, RUNTIME)
    nx.set_node_attributes(tree, {i: random.randint(*memory) for i in range(1, nodes + 1)}, MEMORY)
    for _, _, d in tree.edges(data=True):
        d[RATE] = random.randint(*rate)
        d[DATA] = random.randint(*data)
    tree = nx.relabel_nodes(tree, {0: PLATFORM})
    tree[PLATFORM][1][RATE] = 1
    tree.graph[NAME] = f"random_tree_{time.time()}" if name is None else name
    return tree



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import operator
import pathlib
import pickle
import warnings
from collections.abc import Generator

import networkx as nx
import numpy as np

from slambuc.alg.service.common import *


def encode_service_tree(tree: nx.DiGraph, root: int = 0, pad_size: int = 0) -> np.ndarray[np.int64]:
    """
    Encode the given service *tree* into an array with size of **5*n** where n is the size of the tree.

    The tree must have the tree structure where the root is PLATFORM and the node IDs are increasing integers from *1*
    to *n*. The array's structure is *[r, S_(n-1), R_n, D_n, T_n, M_n]*, where
        - *r* is the root node of the tree (default is PLATFORM that is converted to node *0*),
        - *S_(n-1) is the Prufer sequence of the tree extended with root node PLATFORM,
        - *D_n, R_n* are the ingress edge attributes (DATA, RATE) and
        - *T_n, M_n* are the node attributes (RUNTIME, MEMORY) of the tree nodes in increasing order from *1* to *n*.

    :param tree:        service tree
    :param root:        root node
    :param pad_size:    padding size for uniform length
    :return:            encoded tree as value arrays
    """
    data_seq = np.fromiter(
        itertools.chain(
            [root, *nx.to_prufer_sequence(nx.relabel_nodes(tree.to_undirected(as_view=True), {PLATFORM: root}))],
            (tree[u][v][DATA] for u, v in sorted(tree.edges, key=operator.itemgetter(1))),
            (tree[u][v][RATE] for u, v in sorted(tree.edges, key=operator.itemgetter(1))),
            (tree.nodes[v][RUNTIME] for _, v in sorted(tree.edges, key=operator.itemgetter(1))),
            (tree.nodes[v][MEMORY] for _, v in sorted(tree.edges, key=operator.itemgetter(1)))),
        dtype=np.int64)
    # Padding each vector to have a uniform length for saving
    pad_width = max(0, pad_size - len(tree) + 1)
    return np.pad(data_seq.reshape((5, -1)), ((0, 0), (0, pad_width)))


def decode_service_tree(tdata: np.ndarray[np.int64]) -> nx.DiGraph:
    """
    Decode and rebuild service tree from value arrays.

    Inverse method of :func:`encode_service_tree`.

    :param tdata:   array values
    :return:        service tree
    """
    if tdata.shape[0] % 5:
        warnings.warn(f"Given data with shape {tdata.shape} is not a valid encoded tree!")
    root, *prufer_seq = np.trim_zeros(tdata[0, :], 'b')
    tree = nx.bfs_tree(nx.from_prufer_sequence(prufer_seq), source=root, sort_neighbors=sorted)
    for u, v in tree.edges:
        tree[u][v][DATA], tree[u][v][RATE], tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY] = tdata[1:, v - 1]
    nx.relabel_nodes(tree, {0: PLATFORM}, copy=False)
    tree.graph[NAME] = "tree_" + "".join(map(str, tdata[0]))
    return tree


def save_trees_to_file(trees: list[nx.DiGraph], file_name: str = "test_trees.npy", padding: int = 0):
    """
    Convert trees into a compact format and save them in a single file.

    :param trees:       list of trees
    :param file_name:   output file name
    :param padding:     padding size
    """
    enc_trees = list(encode_service_tree(t, pad_size=padding) for t in trees)
    if enc_trees:
        np.save(file_name, np.stack(enc_trees))


def get_tree_from_file(file_name: str, tree_num: int) -> nx.DiGraph:
    """
    Load and decode a service tree from the given *file_name* with specific ID *tree_num*.

    :param file_name:   file name
    :param tree_num:    tree ID
    :return:            loaded tree
    """
    np_trees = np.load(file_name, mmap_mode="r", allow_pickle=False)
    return decode_service_tree(np_trees[tree_num - 1, :])


def iload_trees_from_file(file_name: str) -> Generator[nx.DiGraph]:
    """
    Generator of service trees loaded from given *file_name*.

    :param file_name:   tree file
    :return:            generator of trees
    """
    np_trees = np.load(file_name, mmap_mode="r", allow_pickle=False)
    for idx in range(np_trees.shape[0]):
        yield decode_service_tree(np_trees[idx, :])


def load_hist_params(hist_dir: str | pathlib.Path, hist_name: str) -> tuple[list[int | float], list[int | float]]:
    """
    Load pickled attributes from given file.

    :param hist_dir:    directory of histogram attributes
    :param hist_name:   name of the histogram
    :return:            loaded histogram attributes
    """
    with open((pathlib.Path(hist_dir, hist_name).with_suffix('.pkl')).resolve(), 'rb') as f:
        return pickle.load(f, fix_imports=True)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib.resources
import itertools
import pathlib
import random
from collections.abc import Generator

import networkx as nx
import numpy as np
import pandas as pd
import scipy

from slambuc.alg.service.common import *
from slambuc.gen.cluster.syn_job import random_job
from slambuc.gen.io import save_trees_to_file
from slambuc.gen.transform import faasify_dag_by_duplication

# Default sample job/task file of the installed spar package
SAMPLES_DIR = importlib.resources.files("slambuc.gen.cluster").joinpath("samples")
DEF_TASK_CSV = SAMPLES_DIR / "sample_tasks.csv"
DEF_BATCH_CSV = SAMPLES_DIR / "batch_task.csv"
DEF_TASK_CSV_HEADER = ('job', 'task', 'duration', 'cpu', 'mem', 'num')
DEF_TASK_CSV_COLS = (1, 2, 3, 4, 5, 6)
# Default attributes for the front-end dispatcher function
DISP_NODE = 0
DISP_RUNTIME = 1
DISP_MEM = 0.1
# Default attributes of tree nodes
DEF_MEM_MAX = 100
DEF_DATA_FACTOR = 0.2

DEF_JOB_TREE_PREFIX = "job_tree"


def convert_tasks_to_dag(job_name: str, tasks: pd.DataFrame, mem_max: int = DEF_MEM_MAX,
                         data_mean: int = None) -> tuple[nx.DiGraph, int]:
    """
    Convert the task lines of given job *job_name* into a DAG and return it with the generated front-end root node.

    :param job_name:    job name in the dataset
    :param tasks:       tasks imported from the dataset
    :param mem_max:     maximum memory to convert memory usage into MB
    :param data_mean:   dataset overhead mean value for generating artificial values
    :return:            generated DAG and dispatcher node
    """
    dag = nx.DiGraph(**{NAME: job_name})
    # Build task DAG
    for _, task in tasks.iterrows():
        # Extract dependencies
        v, *pred = task.task.split('_')
        try:
            v = int(v[1:])
        except ValueError:
            return
        # Add given node and its predecessors
        dag.add_node(v, **{RUNTIME: int(task.duration), CPU: task.cpu / 100, MEMORY: int(task.mem * mem_max)})
        for p in pred:
            dag.add_edge(int(p), v, **{RATE: int(task.num)})
    # Add dispatcher node
    for v in list(filter(lambda _v: not len(dag.pred[_v]), dag)):
        dag.add_edge(DISP_NODE, v, **{RATE: 1})
    dag.nodes[DISP_NODE].update(**{RUNTIME: DISP_RUNTIME, MEMORY: DISP_MEM})
    # Connect dispatcher function to PLATFORM
    dag.add_edge(PLATFORM, DISP_NODE, **{RATE: 1})
    # Artificially fill data overheads
    data_mean = data_mean if data_mean else DEF_DATA_FACTOR * tasks["duration"].mean()
    # Draw data overhead from exponential distribution with lambda = 1 / avg(runtimes) rounded to integer (min 1)
    data_values = dict(zip(dag.edges, np.ceil(scipy.stats.expon(scale=data_mean).rvs(size=len(dag.edges))).astype(int)))
    nx.set_edge_attributes(dag, values=data_values, name=DATA)
    return dag, DISP_NODE


def igenerate_job_tree(job_df: pd.DataFrame, min_size: int = 0) -> Generator[nx.DiGraph]:
    """
    Generate job service trees one-by-one from *min_size*.

    :param job_df:      imported job dataset
    :param min_size:    minimum tree size
    :return:            generator of job DAGs
    """
    jobs = job_df.groupby("job")["task"].count()
    viable_jobs = jobs[min_size <= jobs]
    for job in viable_jobs.index:
        dag, root = convert_tasks_to_dag(job, job_df[job_df["job"] == job])
        if dag is not None:
            yield faasify_dag_by_duplication(dag, root)


def igenerate_syn_tree(n: int | tuple[int, int], iteration: int = 1, job_lb: int = 10) -> Generator[nx.DiGraph]:
    """
    Generate random job service trees based on empirical distributions.

    :param n:           tree size interval
    :param iteration:   number of trees
    :param job_lb:      minimum tree size
    :return:            generator of tree DAGs
    """
    tree_cntr = i = 0
    while tree_cntr < iteration:
        size = random.randint(job_lb, n[1]) if isinstance(n, tuple) else n
        job_df = random_job(task_num=size)
        dag, root = convert_tasks_to_dag(f"syn_job_{i}", job_df)
        tree = faasify_dag_by_duplication(dag, root)
        if ((isinstance(n, int) and len(tree) - 1 == size) or
                (isinstance(n, tuple) and n[0] <= (len(tree) - 1) <= n[1])):
            if tree_cntr and tree_cntr % 10 == 0:
                print(f"Found {tree_cntr} tree[{n=}] out of {i:5d} generated job")
            yield tree
            tree_cntr += 1
        i += 1


########################################################################################################################


def generate_all_job_trees(data_dir: str, task_file: str = DEF_TASK_CSV, start: int = 10, end: int = None,
                           step: int = 10, tree_name: str = DEF_JOB_TREE_PREFIX):
    """
    Generate all job service trees with size interval between *start* and *end* and save them into separate files.

    :param data_dir:    data directory
    :param task_file:   task file name
    :param start:       lower bound of size intervals
    :param end:         upper bound of size intervals
    :param step:        step size of intervals
    :param tree_name:   prefix name of tree files
    """
    print(f"Load data from {task_file}...")
    job_df = pd.read_csv(task_file, usecols=DEF_TASK_CSV_COLS, names=DEF_TASK_CSV_HEADER)
    trees = [tree for tree in igenerate_job_tree(job_df, min_size=start)]
    max_size = max(map(len, trees)) - 1
    end = end if end is not None else max_size
    print(f"Generated {len(trees)} job trees with {start} <= size <= {max_size}")
    for min_size, max_size in itertools.pairwise(range(start, end + step, step)):
        filtered_trees = [t for t in trees if min_size <= len(t) - 1 < max_size]
        if len(filtered_trees):
            print(f"Filtered {len(filtered_trees)} trees with {min_size} <= size <= {max_size}")
            file_name = pathlib.Path(data_dir, f"{tree_name}_n{min_size}-{max_size}.npy").resolve()
            print(f"Saving trees into {file_name}...")
            save_trees_to_file(filtered_trees, file_name=file_name, padding=max_size)
    print("Finished")


def generate_syn_job_trees(data_dir: str, iteration: int = 100, start: int = 10, end: int = 100, step: int = 10,
                           tree_name: str = DEF_JOB_TREE_PREFIX):
    """
    Generate synthetic job service trees with size interval between *start* and *end* and save to separate files
    using extracted empirical distributions.

    :param data_dir:    data directory
    :param iteration:   number of generated trees
    :param start:       lower bound of size intervals
    :param end:         upper bound of size intervals
    :param step:        step size of intervals
    :param tree_name:   prefix name of tree files
    """
    for n in range(start, end + 1, step):
        print(f"Generating synthetic task service trees for {n=} by exhaustive iteration...")
        file_name = pathlib.Path(data_dir, f"syn_{tree_name}_n{n}.npy").resolve()
        output_trees = list(igenerate_syn_tree(n, iteration, job_lb=int(n * 0.5)))
        print(f"Saving trees into {file_name}...")
        save_trees_to_file(output_trees, file_name, padding=n)
    print("Finished")


def generate_mixed_job_trees(data_dir: str, task_file: str = DEF_TASK_CSV, iteration: int = 100, start: int = 10,
                             end: int = 100, step: int = 10, tree_name: str = DEF_JOB_TREE_PREFIX):
    """
    Generate job trees from sample data and extend it with synthetic trees if necessary.

    :param data_dir:    data directory
    :param task_file:   task file name
    :param iteration:   number of generated trees
    :param start:       minimum of size intervals
    :param end:         maximum of size intervals
    :param step:        step size of intervals
    :param tree_name:   prefix name of tree files
    """
    print(f"Load data from {task_file}...")
    job_df = pd.read_csv(task_file, usecols=DEF_TASK_CSV_COLS, names=DEF_TASK_CSV_HEADER)
    trees = [tree for tree in igenerate_job_tree(job_df, min_size=start)]
    max_size = max(map(len, trees)) - 1
    end = end if end is not None else max_size
    print(f"Generated {len(trees)} job trees with {start} <= size <= {max_size}...")
    for min_size, max_size in itertools.pairwise(range(start, end + step, step)):
        filtered_trees = [t for t in trees if min_size <= len(t) - 1 < max_size]
        print(f"Filtered {len(filtered_trees)} trees with {min_size} <= size <= {max_size}")
        if len(filtered_trees) > iteration:
            # Under-sample
            print(f"Under-sampling filtered trees....")
            tree_output = random.sample(filtered_trees, iteration)
        else:
            # Extend with synthetic trees
            syn_tree_num = iteration - len(filtered_trees)
            print(f"Generate additional {syn_tree_num} synthetic job tree...")
            tree_output = list(igenerate_syn_tree((start, max_size), syn_tree_num))
            tree_output = itertools.chain(filtered_trees, tree_output)
        file_name = pathlib.Path(data_dir, f"{tree_name}_n{min_size}-{max_size}.npy").resolve()
        print(f"Saving trees into {file_name}...")
        save_trees_to_file(tree_output, file_name=file_name, padding=max_size)
    print("Finished")


if __name__ == '__main__':
    # generate_all_job_tree("../../../validation/data")
    # generate_syn_job_trees("../../../validation/data")
    generate_mixed_job_trees("../../../validation/data")



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib

import setuptools

setuptools.setup(
    name="SLAMBUC",
    # version=slambuc.__version__,
    description="Serverless Layout Adaptation with Memory-Bounds and User Constraints",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type='text/markdown',
    author="Janos Czentye",
    author_email="czentye@tmit.bme.hu",
    project_urls={"Repository": "https://github.com/hsnlab/SLAMBUC",
                  "Homepage": "https://github.com/hsnlab/SLAMBUC/wiki",
                  "Issue Tracker": "https://github.com/hsnlab/SLAMBUC/issues"},
    packages=setuptools.find_packages(include=['slambuc', 'slambuc.*'],
                                      exclude=['tests', 'validation']),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires='>=3.10',
    license="Apache 2.0",
    keywords="cloud serverless ilp dp tree",
    install_requires=[
        'networkx~=3.3',
        'numpy~=1.26.4',
        'PuLP~=2.9.0',
        'matplotlib~=3.9.1',
        'pandas~=2.2.2',
        'scipy~=1.14.0',
        'cspy~=1.0.3',
    ],
    extras_require={'tests': ['pytest~=8.2.2',
                              'pygraphviz~=1.13',
                              'tabulate~=0.9.0',
                              'docplex~=2.27.239'],
                    'validation': ['tabulate~=0.9.0',
                                   'Click~=8.1.7',
                                   'psutil~=6.0.0']},
    package_data={'*': ['*.pkl',
                        '*.csv']},
    include_package_data=True,
    zip_safe=False
)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import pathlib
import pprint
import random

import networkx as nx

from slambuc.alg.service import PLATFORM, RUNTIME, DATA
from slambuc.alg.util import recreate_subtree_blocks, split_chain, recreate_subchain_blocks
from slambuc.gen.io import encode_service_tree, decode_service_tree, save_trees_to_file, iload_trees_from_file
from slambuc.gen.transform import transform_autonomous_caching
from slambuc.misc.generator import get_random_chain, get_random_tree
from slambuc.misc.plot import draw_tree
from slambuc.misc.util import print_tree_summary, is_compatible


def test_chain_plotter():
    chain = get_random_chain()
    print_tree_summary(chain)
    barr = [1] + sorted(random.sample(range(2, len(chain)), len(chain) // 2))
    partition = split_chain(barr, len(chain))
    print("Partition", partition)
    draw_tree(chain, partition, draw_blocks=True, draw_weights=True)


def test_chain_tree_plotter():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    print_tree_summary(tree)
    barr = [1, 2, 6, 7, 9]
    partition = recreate_subchain_blocks(tree, barr)
    print("Partition", partition)
    draw_tree(tree, partition, draw_blocks=True, draw_weights=True)


def test_random_tree_plotter():
    tree = get_random_tree()
    print_tree_summary(tree)
    barr = random.sample(list(n for n in tree.nodes if n is not PLATFORM), len(tree) // 3)
    if 1 not in barr:
        barr.append(1)
    partition = recreate_subtree_blocks(tree, barr)
    print("Partition", partition)
    draw_tree(tree, partition, draw_blocks=True, draw_weights=True)


def draw_tree_from_file(file_name: str, draw_weights: bool = False):
    draw_tree(nx.read_gml(file_name, destringizer=int), draw_weights=draw_weights)


def test_tree_enc_dec(n: int = 10):
    tree = get_random_tree(n)
    print_tree_summary(tree)
    draw_tree(tree)
    print("Encoded tree:")
    tdata = encode_service_tree(tree)
    print(tdata)
    for part, (i, j) in zip(("Structure", "Data", "Rate", "Runtime", "Memory"),
                            itertools.pairwise(range(0, len(tdata) + 1, n))):
        print(part, tdata[i:j])
    tree2 = decode_service_tree(tdata)
    print("Decoded tree:")
    print_tree_summary(tree2)
    draw_tree(tree2)
    print(f"Isomorphic: {nx.is_isomorphic(tree, tree2)}")
    print(f"Compatible: {is_compatible(tree, tree2)}")


def test_tree_io():
    trees = [get_random_tree(10) for _ in range(10)]
    pprint.pprint("Generated trees:")
    pprint.pprint(trees)
    print("Saving to file...")
    save_trees_to_file(trees, "test_trees.npy")
    print("Loading trees separately from file...")
    for i, tree in enumerate(iload_trees_from_file("test_trees.npy")):
        print(i, "->", tree)
        print("\tisomorphic:", nx.is_isomorphic(trees[i], tree), "compatible:", is_compatible(trees[i], tree))


def test_cache_transform():
    tree = get_random_tree(10)
    print("Generated tree:")
    print(RUNTIME, pprint.pformat(nx.get_node_attributes(tree, name=RUNTIME)))
    print(DATA, pprint.pformat(nx.get_edge_attributes(tree, name=DATA)))
    tree2 = transform_autonomous_caching(tree, 1)
    print("Transformed tree:")
    print(RUNTIME, pprint.pformat(nx.get_node_attributes(tree2, name=RUNTIME)))


if __name__ == '__main__':
    # test_chain_plotter()
    # test_chain_tree_plotter()
    # test_random_tree_plotter()
    # draw_tree_from_file(pathlib.Path(__file__).parent / "data/graph_test_tree_ser_latency1.gml", draw_weights=False)
    # draw_tree_from_file(pathlib.Path(__file__).parent / "data/graph_test_tree_ser_latency2.gml", draw_weights=True)
    # draw_tree_from_file(pathlib.Path(__file__).parent / "data/graph_test_tree_par_btree.gml", draw_weights=False)
    # draw_tree_from_file(pathlib.Path(__file__).parent / "data/graph_test_tree_par_ltree.gml", draw_weights=False)
    # test_tree_enc_dec()
    # test_tree_io()
    test_cache_transform()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import math

import pulp as lp

from slambuc.alg import LP_LAT, INFEASIBLE, T_IBLOCK_GEN, T_RESULTS, T_PART
from slambuc.alg.util import (ser_block_sublatency, split_chain, ser_block_subcost, ser_block_submemory,
                              iser_mul_factor)


def ifeasible_greedy_blocks(memory: list[int], M: int) -> T_IBLOCK_GEN:
    """
    Generate all feasible (connected) blocks that meet the memory constraint *M* in a greedy manner.

    Block memories are calculated assuming serialized platform execution model.

    :param memory:  list of node memory values
    :param M:       upper block memory limit
    :return:        generator of blocks
    """
    n = len(memory)
    yield from ((i, j) for i in range(n) for j in range(i, n) if ser_block_submemory(memory, i, j) <= M)


def ifeasible_blocks(memory: list[int], M: int) -> T_IBLOCK_GEN:
    """
    Generate all feasible (connected) blocks that meet the memory constraint *M* assuming serialized executions.

    :param memory:  list of node memory values
    :param M:       upper block memory limit
    :return:        generator of blocks
    """
    n = len(memory)
    for i in range(n):
        cumsum = 0
        for j in range(i, n):
            if (cumsum := cumsum + memory[j]) > M:
                break
            else:
                yield i, j


def build_chain_cfg_model(runtime: list[int], memory: list[int], rate: list[int], data: list[int],
                          M: int = math.inf, L: int = math.inf, start: int = 0, end: int = None,
                          delay: int = 1) -> tuple[lp.LpProblem, dict[lp.LpVariable]]:
    """
    Generate the configuration ILP model for chains.

    Block metrics are calculated assuming serialized platform execution model.

    :return: tuple of the created LP model and the dict of created decision variables
    """
    # Model
    model = lp.LpProblem(name="Chain_Partitioning", sense=lp.LpMinimize)
    # Decision variables
    X = {(b, w): lp.LpVariable(f'x_{b}_{w}', cat=lp.LpBinary) for b, w in ifeasible_blocks(memory, M)}
    # Objective
    model += lp.lpSum(ser_block_subcost(runtime, rate, data, b, w) * X[b, w] for b, w in X)
    # Feasibility constraints
    for j in range(len(runtime)):
        model += lp.lpSum(X[b, w] for b, w in X if b <= j <= w) == 1, f"C_{j:03d}"
    # Latency constraint
    sum_lat = lp.lpSum(ser_block_sublatency(runtime, rate, data, b, w, delay, start, end) * X[b, w] for b, w in X)
    if L < math.inf:
        model += sum_lat <= L, LP_LAT
    else:
        # Add redundant constraint to implicitly calculate the latency value
        model += sum_lat >= 0, LP_LAT
    return model, X


def chain_cfg_partitioning(runtime: list[int], memory: list[int], rate: list[int], data: list[int],
                           M: int = math.inf, L: int = math.inf, start: int = 0, end: int = None,
                           delay: int = 1, solver: lp.LpSolver = None) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a chain based on the configuration ILP formalization.

    Block metrics are calculated assuming serialized platform execution model.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param data:    input data fetching delay in ms
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param start:   head node of the latency-limited subchain
    :param end:     tail node of the latency-limited subchain
    :param delay:   invocation delay between blocks
    :param solver:  specific solver class (default: COIN-OR CBC)
    :return:        tuple of partitioning blocks, optimal cost, and the calculated latency of the subchain
    """
    model, X = build_chain_cfg_model(runtime, memory, rate, data, M, L, start, end, delay)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = lp.value(model.objective), lp.value(model.constraints[LP_LAT])
        return extract_blocks_from_xvars(X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def recreate_blocks_from_xvars(X: dict[tuple[int, int], lp.LpVariable], n: int) -> T_PART:
    """
    Extract barrier nodes from variable dict and recreate partitioning blocks.

    :param X:   dict of decision variables
    :param n:   chain size
    :return:    partition blocks
    """
    return split_chain(barr=[b for (b, _), x in X.items() if x.varValue == 1], n=n)


def extract_blocks_from_xvars(X: dict[tuple[int, int], lp.LpVariable]) -> T_PART:
    """
    Extract interval boundaries [b, w] relying on the decision variable's structure.

    :param X:   dict of decision variables
    :return:    partition blocks
    """
    return [list(range(b, w + 1)) for (b, w), x in X.items() if x.varValue == 1]


########################################################################################################################


def build_greedy_chain_mtx_model(runtime: list[int], memory: list[int], rate: list[int],
                                 data: list[int], M: int = math.inf, L: int = math.inf,
                                 delay: int = 1) -> tuple[lp.LpProblem, list[list[lp.LpVariable]]]:
    """
    Generate the matrix ILP model for chains by calculating block metrics greedily using utility functions.

    Block metrics are calculated assuming serialized platform execution model.

    :return: tuple of the created LP model and the dict of created decision variables
    """
    # Model
    model = lp.LpProblem(name="Chain_Partitioning", sense=lp.LpMinimize)
    n = len(runtime)
    # Decision variable matrix
    X = [[lp.LpVariable(f"x_{i}_{j}", cat=lp.LpBinary) for j in range(i + 1)] for i in range(n)]
    # Objective
    sum_cost = lp.LpAffineExpression()
    for j in range(n):
        cost_pre = 0
        for i in range(j, n):
            cost_ji = ser_block_subcost(runtime, rate, data, j, i)
            sum_cost += (cost_ji - cost_pre) * X[i][j]
            cost_pre = cost_ji
    model += sum_cost
    # Feasibility constraints
    for i in range(n):
        model += lp.lpSum(X[i]) == 1, f"C_f{i:02d}"
    # Knapsack constraints
    for j in range(n):
        model += lp.lpSum(memory[i] * X[i][j] for i in range(j, n)) <= M, f"C_k{j:02d}"
    # Connectivity constraints
    for j in range(n):
        for i in range(j + 1, n):
            model += X[i - 1][j] - X[i][j] >= 0, f"C_c{j}_{i}"
    # Latency constraint
    sum_lat = lp.LpAffineExpression()
    for j in range(n):
        lat_pre = 0
        for i in range(j, n):
            lat_ji = ser_block_sublatency(runtime, rate, data, j, i, delay, 0, n - 1)
            sum_lat += (lat_ji - lat_pre) * X[i][j]
            lat_pre = lat_ji
    if L < math.inf:
        model += sum_lat <= L, LP_LAT
    else:
        # Add redundant constraint to implicitly calculate the latency value
        model += sum_lat >= 0, LP_LAT
    return model, X


def build_chain_mtx_model(runtime: list[int], memory: list[int], rate: list[int],
                          data: list[int], M: int = math.inf, L: int = math.inf,
                          delay: int = 1) -> tuple[lp.LpProblem, list[list[lp.LpVariable]]]:
    """
    Generate the matrix ILP model for chains.

    Block metrics are calculated assuming serialized platform execution model.

    :return: tuple of the created model and list of decision variables
    """
    # Model
    model = lp.LpProblem(name="Chain_Partitioning", sense=lp.LpMinimize)
    n = len(runtime)
    # Decision variable matrix
    X = [[lp.LpVariable(f"x_{i}_{j}", cat=lp.LpBinary) for j in range(i + 1)] for i in range(n)]
    # Objective
    model += lp.lpSum(lp.lpSum((rate[j] * data[j] * X[j][j] if i == j else  # Add data fetching to first/barrier node
                                -1 * rate[i] * data[i] * X[i][j],  # Remove caching of prior last node
                                rate[i] * runtime[i] * X[i][j],  # Add runtime cost
                                rate[i + 1] * data[i + 1] * X[i][j] if i < n - 1 else 0))  # Add data caching
                      for j in range(n) for i in range(j, n))
    # Feasibility constraints
    for i in range(n):
        model += lp.lpSum(X[i]) == 1, f"C_f{i:02d}"
    # Knapsack constraints
    for j in range(n):
        model += lp.lpSum(memory[i] * X[i][j] for i in range(j, n)) <= M, f"C_k{j:02d}"
    # Connectivity constraints
    for j in range(n):
        for i in range(j + 1, n):
            model += X[i - 1][j] - X[i][j] >= 0, f"C_c{j}_{i}"
    # Latency constraint
    sum_lat = lp.LpAffineExpression()
    for j in range(n):
        n_pre = 1
        for i, n_i in enumerate(iser_mul_factor(itertools.islice(rate, j, n)), start=j):
            if i == j:
                # Add data fetching
                sum_lat += n_i * data[j] * X[j][j]
                if j > 0:
                    # Add block invocation delay
                    sum_lat += delay * X[j][j]
            else:
                # Remove prior last node data caching
                sum_lat -= n_pre * data[i] * X[i][j]
            # Add function instances
            sum_lat += n_i * runtime[i] * X[i][j]
            if i < n - 1:
                # Add caching
                sum_lat += n_i * data[i + 1] * X[i][j]
            n_pre = n_i
    if L < math.inf:
        model += sum_lat <= L, LP_LAT
    else:
        # Add redundant constraint to implicitly calculate the latency value
        model += sum_lat >= 0, LP_LAT
    return model, X


def chain_mtx_partitioning(runtime: list[int], memory: list[int], rate: list[int], data: list[int],
                           M: int = math.inf, L: int = math.inf, delay: int = 1, solver: lp.LpSolver = None,
                           **kwargs) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a chain based on the matrix ILP formalization.

    Block metrics are calculated assuming serialized platform execution model.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param data:    input data fetching delay in ms
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param delay:   invocation delay between blocks
    :param solver:  specific solver class (default: COIN-OR CBC)
    :return:        tuple of partitioning blocks, optimal cost, and the calculated latency of the subchain
    """
    model, X = build_chain_mtx_model(runtime, memory, rate, data, M, L, delay)
    status = model.solve(solver=solver if solver else lp.PULP_CBC_CMD(mip=True, msg=False))
    if status == lp.LpStatusOptimal:
        opt_cost, opt_lat = lp.value(model.objective), lp.value(model.constraints[LP_LAT])
        return recreate_blocks_from_xmatrix(X), opt_cost, L + opt_lat if L < math.inf else opt_lat
    else:
        return INFEASIBLE


def recreate_blocks_from_xmatrix(X: list[list[lp.LpVariable]]) -> T_PART:
    """
    Extract barrier nodes from decision variable matrix and recreate partitioning blocks.

    :param X:   matrix of decision variables
    :return:    partition blocks
    """
    return split_chain(barr=list(filter(lambda i: X[i][i].value(), range(len(X)))), n=len(X))


def extract_blocks_from_xmatrix(X: list[list[lp.LpVariable]]) -> T_PART:
    """
    Extract interval boundaries [b, w] directly from decision variable matrix.

    :param X:   matrix of decision variables
    :return:    partition blocks
    """
    return [list(itertools.takewhile(lambda i: X[i][j].value(), range(j, len(X))))
            for j in range(len(X)) if X[j][j].value()]



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
import math
import time
from collections.abc import Generator, Callable

import cspy
import networkx as nx
import numpy as np

from slambuc.alg import INFEASIBLE, T_BLOCK, T_RESULTS, T_PART
from slambuc.alg.service.common import SEP, ASSIGN, Flavor
from slambuc.alg.util import (leaf_label_nodes, ibacktrack_chain, iflattened_tree, gen_subtree_memory, gen_subtree_cost,
                              gen_subchain_latency, verify_limits)

# Naming convention for state-space DAG required by *cspy* lib
START, END = 'Source', 'Sink'
COST, LAT = 'weight', 'res_cost'


def encode_state(grp: T_BLOCK, flavor: Flavor) -> str:
    """
    Encode DAG node name with flavor's memory as a unique str (hashable).

    :param grp:     partition block
    :param flavor:  assigned *flavor*
    :return:        encoded partition block
    """
    return f"{SEP.join(map(str, grp))}{ASSIGN}{SEP.join(map(str, flavor))}"


def decode_state(name: str) -> list[T_BLOCK, str]:
    """
    Decode DAG node name from encoded str into partition block (list of int) and flavor's memory (mem).

    :param name:    encoded partition block
    :return:        decoded block and assigned flavor
    """
    parts = name.rsplit(ASSIGN, maxsplit=1)
    return list(map(int, parts[0].split(SEP))), Flavor(*parts[1].split(SEP)).name


def ibuild_gen_csp_dag(tree: nx.DiGraph, root: int = 1, flavors: list[Flavor] = (Flavor(),),
                       exec_calc: Callable[[int, int, int], int] = lambda i, t, n: t,
                       cpath: set[int] = frozenset(), delay: int = 1) -> Generator[nx.DiGraph, list[list[int]]]:
    """
    Calculate all state-space DAGs of the given *tree* based on the alternative chain decompositions.

    The given flavors as list of (memory, CPU, cost_factor) tuples define the available memory (and group upper limit),
    available relative vCPU cores and relative cost multiplier.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rate
    :param root:        root node of the graph
    :param flavors:     list of flavors resources given by the tuple of available *(memory, relative CPU cores)*
    :param exec_calc:   function that calculates the effective runtimes from reference runtime and available CPU cores
    :param cpath:       critical path in the form of subchain[root -> cp_end]
    :param delay:       invocation delay between blocks
    :return:            generated DAG graph and the related nodes of the flattened tree
    """
    # Annotate nodes with reachable leafs of tree
    tree = leaf_label_nodes(tree)
    # Iterate over the feasible chain-flattened tree
    for chains in iflattened_tree(tree, root):
        # Initiate data structure for DAG
        _cache = collections.defaultdict(list)
        _cache.update({START: [START], END: [END]})
        dag = nx.DiGraph(directed=True, n_res=2, **tree.graph)
        # Iterate the subchains backward
        for prev, chain in itertools.pairwise(itertools.chain([_cache[END]], reversed(chains))):
            # Generate the subcases of the given subchain
            for i, b in enumerate(reversed(chain), start=1):
                grp = []
                for j, v in enumerate(chain[-i:], start=1):
                    grp = grp + [v]
                    for f in flavors:
                        # Skip infeasible subcase due to memory constraint
                        if gen_subtree_memory(tree, b, grp, f.ncore) > f.mem:
                            break
                        grp_id = encode_state(grp, f)
                        _cache[b].append(grp_id)
                        # Get starting node of dependent subcases from the cache
                        nxt = prev[0] if v == chain[-1] else chain[-i + j]
                        # Calculate subcase cost and latency
                        grp_cost = round(gen_subtree_cost(tree, b, grp, f.ncore, exec_calc) * f.cfactor)
                        grp_lat = gen_subchain_latency(tree, b, set(grp), cpath, f.ncore, exec_calc)
                        # Add invocation delay for inter-block invocations
                        if b != root and grp_lat > 0:
                            grp_lat += delay
                        # Add connection between related subcases
                        for sc in _cache[nxt]:
                            dag.add_edge(grp_id, sc, **{COST: grp_cost, LAT: np.array([1, grp_lat])})
            # Remove unnecessary subcases from cache
            for p in prev:
                del _cache[p]
        # Add initial connections from START node
        for sc in _cache[root]:
            dag.add_edge(START, sc, **{COST: 0, LAT: np.array([1, 0])})
        yield dag, chains


def csp_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                          N: int = 1, cp_end: int = None, delay: int = 1, exhaustive: bool = True,
                          solver=cspy.BiDirectional, timeout: int = None, **cspargs) -> T_RESULTS:
    """
    Calculate minimal-cost partitioning of a *tree* based on constrained shortest path (CSP) formalization
    without considering flavor assignment.

    Details in: T. Elgamal at al.: “Costless: Optimizing Cost of Serverless Computing through Function Fusion
    and Placement,” in 2018 IEEE/ACM Symposium on Edge Computing (SEC), 2018, pp. 300–312. doi: 10.1109/SEC.2018.00029.

    :param tree:        service tree annotated with node runtime(ms), memory(MB) and edge rate
    :param root:        root node of the graph
    :param M:           upper memory bound of the partition blocks in MB
    :param L:           latency limit defined on the critical path in ms
    :param N:           available CPU core count
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:       invocation delay between blocks
    :param exhaustive:  iterate over all topological ordering of the service tree or stop at first feasible solution
    :param solver:      specific solver class (default: cspy.BiDirectional)
    :param timeout:     time limit in sec
    :param cspargs:     additional CSP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    best_res, t_start = INFEASIBLE, time.process_time()
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    for dag, chains in ibuild_gen_csp_dag(tree, root, [Flavor(M, N, 1.0)], cpath=cpath, delay=delay):
        try:
            model = solver(dag, max_res=[len(dag.edges), L], min_res=[1, 0], time_limit=timeout, **cspargs)
            model.run()
            if model.path is not None and model.total_cost < best_res[1]:
                best_res = extract_grp_from_path(model.path, flavors=False), model.total_cost, model.consumed_resources[
                    -1]
                if not exhaustive:
                    # Stop at first feasible solution
                    break
        except Exception:
            # No feasible solution
            continue
        finally:
            if timeout and time.process_time() - t_start > timeout:
                break
    return best_res


def csp_gen_tree_partitioning(tree: nx.DiGraph, root: int = 1, flavors: list[tuple[int, int]] = ((math.inf, 1),),
                              exec_calc: collections.abc.Callable[[int, int], int] = lambda i, t, n: t,
                              L: int = math.inf, cp_end: int = None, delay: int = 1, solver=cspy.BiDirectional,
                              timeout: int = None, **cspargs) -> T_RESULTS:
    """
    Calculate minimal-cost partitioning of a *tree* based on constrained shortest path (CSP) formalization with
    incorporated flavor assignment.

    Details in: T. Elgamal at al.: “Costless: Optimizing Cost of Serverless Computing through Function Fusion
    and Placement,” in 2018 IEEE/ACM Symposium on Edge Computing (SEC), 2018, pp. 300–312. doi: 10.1109/SEC.2018.00029.


    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rate
    :param root:        root node of the graph
    :param flavors:     list of flavors resources given by the tuple of available *(memory, rel CPU cores, cost factor)*
    :param exec_calc:   function that calculates the effective runtimes from reference runtime and available CPU cores
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:       invocation delay between blocks
    :param solver:      specific solver class (default: cspy.BiDirectional)
    :param timeout:     time limit in sec
    :param cspargs:     additional CSP solver parameters
    :return:            tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    best_res = INFEASIBLE
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, max(m for m, *_ in flavors), L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    for dag, chains in ibuild_gen_csp_dag(tree, root, flavors, exec_calc, cpath, delay):
        model = solver(dag, max_res=[len(dag.edges), L], min_res=[1, 0], time_limit=timeout, **cspargs)
        try:
            model.run()
            if model.path is not None and model.total_cost < best_res[1]:
                best_res = extract_grp_from_path(model.path), model.total_cost, model.consumed_resources[-1]
        except Exception:
            # No feasible solution
            continue
    return best_res


def extract_grp_from_path(path: list[str], flavors: bool = True) -> T_PART:
    """
    Extract partitioning from *path* and recreate partition blocks.

    :param path:    solution path of the CSP graph
    :param flavors: whether return flavors or not
    :return:        resulted partitioning blocks
    """
    return sorted(decode_state(grp) if flavors else decode_state(grp)[0] for grp in path[1:-1])



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
import math
import operator
import typing
import warnings

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.service import *
from slambuc.alg.tree.ser.pseudo import SubLTreePart, OPT
from slambuc.alg.util import (ipostorder_dfs, ileft_right_dfs, ibacktrack_chain, recreate_subtree_blocks,
                              par_inst_count, verify_limits)


class SubParBTreePart(typing.NamedTuple):
    """Store subtree partitioning attributes for a given subcase."""
    cost: int = math.inf  # Optimal sum cost of the subtree partitioning (OPT)
    top_cost: int = math.inf  # Cost of the topmost subtree block
    top_blk: set[int] = set()  # Nodes of the topmost block
    barr: set[int] = set()  # Barrier/heading nodes of the given subtree partitioning

    def __repr__(self):
        return repr(tuple(self))


def pseudo_par_btree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf, N: int = 1,
                                  cp_end: int = None, delay: int = 1, bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying bottom-up tree traversal approach.

    Block metrics are calculated based on parallelized execution platform model.

    Provide suboptimal partitioning due to the inaccurate latency calculation that directly comes from the bottom-up
    tree traversal approach.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param N:               available CPU core count
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    if N > 1:
        warnings.warn(f"Suboptimal for {N=} due to violated monotonicity of cost function in upward merging!")
    # Set of critical path's nodes
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # Init empty data structure for optimal results of subtrees T_v
    DP = {n: collections.defaultdict(dict) for n in tree if n is not PLATFORM}
    # Iterate nodes in a bottom-up traversal order
    for p, v in ipostorder_dfs(tree, root):
        r_v, d_v, t_v, m_v = tree[p][v][RATE], tree[p][v][DATA], tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY]
        # SINGLETON: calculate the default subcase of singleton partition of node v
        cost = r_v * (d_v + t_v) + sum(par_inst_count(r_v, vs[RATE], N) * vs[DATA] for vs in tree.succ[v].values())
        # Only add fetching overhead for root node and omit caching to ensure lat monotonicity of merging
        lat = d_v + t_v if v == root else t_v if v in cpath else 0
        DP[v][lat, lat][m_v, m_v] = SubParBTreePart(cost, cost, {v}, {v})
        # Bottom-up propagation for considering v's descendant subcases in sequential order (v is not leaf)
        # -->   p   ---->   v   ----[i.]---->   b
        for b in tree.successors(v):
            # Init empty data structure of subcase T[v,b]
            _cache = collections.defaultdict(dict)
            r_b, d_b, t_b = tree[v][b][RATE], tree[v][b][DATA], tree.nodes[b][RUNTIME]
            # Calculate possible latency/memory combinations
            for lats_v, lats_b in itertools.product(DP[v], DP[b]):
                for mem_v, mem_b in itertools.product(DP[v][lats_v], DP[b][lats_b]):
                    sub_v, sub_b = DP[v][lats_v][mem_v], DP[b][lats_b][mem_b]
                    lat_v, blk_lat_v = lats_v
                    lat_b, blk_lat_b = lats_b
                    # Latency calculation in case v -> b edge is in cpath
                    if b in cpath:
                        # CUT: v -> b edge is marked as a cut (opt subcases of b have the dedicated memory value 0)
                        if mem_b == (OPT, OPT):
                            # Add caching overhead of inter-block call
                            top_blk_lat = blk_lat_v + math.ceil(r_b / (r_v * N)) * d_b
                            # Add invocation delay and fetch overhead of inter-block call
                            if (lat := top_blk_lat + delay + d_b + lat_b) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                        # MERGE: v -> b edge is marked as a merge
                        else:
                            # Top block's latency comes for only the single cp node v, hence n_v = 1
                            top_blk_lat = blk_lat_v + math.ceil(r_b / (r_v * N)) * blk_lat_b
                            # Calculate new latency based on the recalculated top block's latency
                            if (lat := lat_b - blk_lat_b + top_blk_lat) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                    else:
                        # Top block's attributes remain the same, either v in cpath or not
                        top_blk_lat, lat = blk_lat_v, lat_v
                    # CUT: v -> b edge is marked as a cut
                    if mem_b == (OPT, OPT):
                        # Top block's attributes remain the same
                        mem, top_blk = mem_v, sub_v.top_blk
                        # Sub-partitions are just concatenated
                        barr, cost, top_blk_cost = sub_v.barr | sub_b.barr, sub_v.cost + sub_b.cost, sub_v.top_cost
                    # MERGE: v -> b edge is marked as a merge
                    elif (mem_ld := mem_v[0] + mem_b[0]) > M:
                        # Infeasible subcase due to exceeded memory constraint
                        continue
                    else:
                        mem_op, top_blk_cost = mem_v[1], sub_v.top_cost - par_inst_count(r_v, r_b, N) * d_b
                        for f in sub_b.top_blk:
                            r_f = tree[next(tree.predecessors(f))][f][RATE]
                            if (mem_op := max(mem_op, min(math.ceil(r_f / r_v), N) * tree.nodes[f][MEMORY])) > M:
                                # Infeasible subcase due to exceeded memory constraint
                                continue
                            top_blk_cost += (par_inst_count(r_v, r_f, N) * tree.nodes[f][RUNTIME] +
                                             sum(par_inst_count(r_v, fs[RATE], N) * fs[DATA]
                                                 for s, fs in tree.succ[f].items() if s not in sub_b.top_blk))
                        # Top blocks of the two sub-partitions are merged together with root node v
                        mem = mem_ld, mem_op
                        barr, top_blk = sub_v.barr | sub_b.barr - {b}, sub_v.top_blk | sub_b.top_blk
                        cost = sub_v.cost - sub_v.top_cost + sub_b.cost - sub_b.top_cost + top_blk_cost
                    # Store the min cost subcase
                    sub_lats = lat, top_blk_lat
                    if mem not in _cache[sub_lats] or cost < _cache[sub_lats][mem].cost:
                        _cache[sub_lats][mem] = SubParBTreePart(cost, top_blk_cost, top_blk, barr)
                    # Check whether the current subcase is a better sub-solution for state (v, sub_lats, mem)
                    if mem not in _cache[sub_lats] or cost < _cache[sub_lats][mem].cost:
                        if bidirectional:
                            # Eliminate prior subcases that are dominated by the new subcase
                            for l, m in tuple((_l, _m) for _l in _cache if _l[0] >= lat and _l[1] >= top_blk_lat
                                              for _m in _cache[_l] if (_m[0] >= mem[0] and _m[1] >= mem[1]
                                                                       and _cache[_l][_m].cost >= cost)):
                                if len(_cache[l]) > 1:
                                    del _cache[l][m]
                                else:
                                    del _cache[l]
                        # Add superior subcase
                        _cache[sub_lats][mem] = SubParBTreePart(cost, top_blk_cost, top_blk, barr)
            # Store min subcases as C(v,i-1) for the next iteration of the propagation process
            DP[v] = _cache
            # Drop unnecessary subcases
            del DP[b]
        # Store the cost-opt subcases wrt. latency for node v encoded with memory value 0,0
        for lats in DP[v]:
            DP[v][lats][OPT, OPT] = min(DP[v][lats].values(), key=operator.itemgetter(OPT))
    # Subcases under the root node contain the feasible partitioning
    if opt_lats := min(DP[root], key=lambda _l: DP[root][_l][OPT, OPT].cost, default=None):
        opt = DP[root][opt_lats][OPT, OPT]
        return recreate_subtree_blocks(tree, opt.barr), opt.cost, opt_lats[OPT]
    else:
        # No feasible solution
        return INFEASIBLE


########################################################################################################################


def pseudo_par_ltree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf, N: int = 1,
                                  cp_end: int = None, delay: int = 1, bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying left-right tree traversal approach.

    Block metrics are calculated based on parallelized execution platform model.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param N:               available CPU core count
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Set of critical path's nodes
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # Init empty data structure for optimal results of subtrees
    TDP = {n: {} for n in tree if n is not PLATFORM}
    # Process subtrees of T in a bottom-up traversal order
    for p, n in ipostorder_dfs(tree, root):
        # Init empty data structure for optimal subcases T_n[v,b]
        DP = collections.defaultdict(lambda: collections.defaultdict(dict))
        r_n, d_n, t_n, m_n = tree[p][n][RATE], tree[p][n][DATA], tree.nodes[n][RUNTIME], tree.nodes[n][MEMORY]
        # Traverse subtree T_n in a left-right traversal
        #       |--(j-1.)--> [v_p]   |--(i-1.)--> [b_p]                                 | left side
        # -->  [u]  ----(j.)---->  [v]  ----(i.)---->  [b]  ----(max_k.)----> [b_k]     | right side
        for e_uvp, b_p, e_vb in ileft_right_dfs(tree, n):
            v, b = e_vb
            # Merge edge u -> v as consider node v and all its first 0 child (b=0)
            if b == 0:
                # SINGLETON: C[n,0] - calculate the default subcase of singleton partition of root node n
                if v == n:
                    cost = r_n * (d_n + t_n) + sum(par_inst_count(r_n, ns[RATE], N) * ns[DATA]
                                                   for ns in tree.succ[n].values())
                    if n in cpath:
                        nc = next(filter(lambda c: c in cpath, tree.successors(n)), None)
                        # n_v = 1 for root node n; add caching overhead on critical path here
                        lat = d_n + t_n + (math.ceil(tree[n][nc][RATE] / (r_n * N)) * tree[n][nc][DATA] if nc else 0)
                    else:
                        lat = 0
                    DP[e_vb][lat][m_n, m_n] = SubLTreePart(cost, 1, {n})
                # MERGE: C[v,0] - subcase of merging single node v into the n-rooted block containing node u
                else:
                    u, v_p = e_uvp
                    r_v, d_v = tree[u][v][RATE], tree[u][v][DATA]
                    t_v, m_v = tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY]
                    # v_p = 0 => no prior sibling node, C[v,0] = C[u,j-1] + delta(u,v)
                    for lat_u in DP[e_uvp]:
                        for mem_u in DP[e_uvp][lat_u]:
                            sub_u = DP[e_uvp][lat_u][mem_u]
                            # Calculate memory prefetch and operative memory demands separately
                            mem = mem_u[0] + m_v, max(mem_u[1], min(math.ceil(r_v / r_n), N) * m_v)
                            if max(mem) > M:
                                # Infeasible subcase due to exceeded memory constraint
                                continue
                            # v in cpath => u in cpath
                            if v in cpath:
                                rel_r_v = math.ceil(r_v / (tree[next(tree.predecessors(u))][u][RATE] * N))
                                n_v = rel_r_v * sub_u.mul
                                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                                w_v = math.ceil(tree[v][vc][RATE] / (r_v * N)) * tree[v][vc][DATA] if vc else 0
                                # Recalculate caching overhead (remove prior block's write out from E2E latency)
                                if (lat := lat_u + n_v * (t_v - d_v + w_v)) > L:
                                    # Infeasible subcase due to exceeded latency constraint
                                    continue
                            else:
                                lat, n_v = lat_u, sub_u.mul
                            # Calculate sum cost by adding merged node runtime and caching
                            cost = (sub_u.cost + par_inst_count(r_n, r_v, N) * (t_v - d_v) +
                                    sum(par_inst_count(r_n, vs[RATE], N) * vs[DATA] for vs in tree.succ[v].values()))
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if lat not in DP[e_vb] or mem not in DP[e_vb][lat] or cost < DP[e_vb][lat][mem].cost:
                                # Add superior subcase
                                DP[e_vb][lat][mem] = SubLTreePart(cost, n_v, sub_u.barr)
            # CUT: C[v,b]
            else:
                # b_k <- C[v_i, d(v_i)]
                b_k = collections.deque(tree.successors(b), 1).pop() if len(tree.succ[b]) else 0
                # Previously merged edge: v -> b  =>  [root ... -> u -> v -> b ...]   |   [b]  --(max_k.)--> [b_w]
                DP[e_vb] = DP[b, b_k]
                # Cut edge: v -> b  |  [root ... -> u -> v] - [b ...]   |   v --(i-1.)--> b_p
                for lat_v in DP[v, b_p]:
                    for mem_v in DP[v, b_p][lat_v]:
                        for lat_b, opt_b in TDP[b].items():
                            # b in cpath => v in cpath
                            if (lat := lat_v + delay + lat_b if b in cpath else lat_v) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                            sub_v = DP[v, b_p][lat_v][mem_v]
                            # Only concatenate blocks
                            cost, barr = sub_v.cost + opt_b.cost, sub_v.barr | opt_b.barr
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if lat not in DP[e_vb] or mem_v not in DP[e_vb][lat] or cost < DP[e_vb][lat][mem_v].cost:
                                if bidirectional:
                                    # Eliminate prior subcases that are dominated by the new subcase
                                    for l, m in tuple((_l, _m) for _l in DP[e_vb] if _l >= lat
                                                      for _m in DP[e_vb][_l] if (_m[0] >= mem_v[0] and _m[0] >= mem_v[0]
                                                                                 and DP[e_vb][_l][_m].cost >= cost)):
                                        if len(DP[e_vb][l]) > 1:
                                            del DP[e_vb][l][m]
                                        else:
                                            del DP[e_vb][l]
                                # Add superior subcase
                                DP[e_vb][lat][mem_v] = SubLTreePart(cost, sub_v.mul, barr)
        # Cache the best subcase for subtree T_n
        n_w = collections.deque(tree.successors(n), 1).pop() if len(tree.succ[n]) else 0
        # Store the best subcases of subtree T_n
        for lat_n, dp in DP[n, n_w].items():
            TDP[n][lat_n] = min(dp.values(), key=operator.itemgetter(OPT))
    # Subcases under the root node contain the feasible partitioning
    if opt_lat := min(TDP[root], key=lambda _l: TDP[root][_l].cost, default=None):
        opt = TDP[root][opt_lat]
        return recreate_subtree_blocks(tree, opt.barr), opt.cost, opt_lat
    else:
        # No feasible solution
        return INFEASIBLE



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import pprint
import sys
import time

import networkx as nx
import pulp
import pytest
import tabulate

from slambuc.alg.service import *
from slambuc.alg.tree.ser.ilp import (ifeasible_subtrees, ifeasible_greedy_subtrees, build_tree_cfg_model,
                                      tree_hybrid_partitioning, extract_subtrees_from_xdict,
                                      recreate_subtrees_from_xdict)
from slambuc.alg.util import ibacktrack_chain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import print_lp_desc, evaluate_ser_tree_partitioning, get_cplex_path, get_glpk_path


def test_feasible_subtrees(branch: int = 2, depth: int = 2):
    tree = nx.balanced_tree(branch, depth, create_using=nx.DiGraph)
    nx.set_node_attributes(tree, 1, MEMORY)
    tree.add_edge(PLATFORM, 0)
    print("  All blocks (exhaustive)  ".center(80, '='))
    print(f"{tree = !s}")
    _start = time.perf_counter()
    blocks = list(ifeasible_greedy_subtrees(tree, root=None, M=math.inf))
    _stop = time.perf_counter()
    print(f"Number of generated blocks: {len(blocks)}")
    print(f"Sum time: {(_stop - _start) * 1000} ms")
    print("  All blocks (bottom-up)  ".center(80, '='))
    print(f"{tree = !s}")
    _start = time.perf_counter()
    blocks = list(ifeasible_subtrees(tree, root=0, M=math.inf))
    _stop = time.perf_counter()
    print(f"Number of generated blocks: {len(blocks)}")
    print(f"Sum time: {(_stop - _start) * 1000} ms")
    pprint.pprint(blocks)


def test_restricted_feasible_subtrees():
    memory, M = {i: m for i, m in enumerate((3, 3, 2, 1, 3, 1, 3, 2, 2, 1, 2, 1, 1, 3, 3, 2))}, 6
    print(f"Data: {memory = }, {M = }")
    tree = nx.balanced_tree(2, 3, create_using=nx.DiGraph)
    nx.set_node_attributes(tree, memory, MEMORY)
    tree.add_edge(PLATFORM, 0)

    print("  Restricted blocks (exhaustive)  ".center(80, '='))
    print(f"{tree = !s}")
    _start = time.perf_counter()
    blocks = list(ifeasible_greedy_subtrees(tree, root=None, M=M))
    _stop = time.perf_counter()
    print(f"Number of generated blocks: {len(blocks)}")
    print(f"Sum time: {(_stop - _start) * 1000} ms")
    mem_data = [[memory[v] for v in b] for _, b in blocks]
    greedy_blocks_data = list(zip(blocks, mem_data, map(sum, mem_data)))
    print(tabulate.tabulate(greedy_blocks_data, ('Block', 'Memory', 'Sum')))

    print("  Restricted blocks (bottom-up)  ".center(80, '='))
    print(f"{tree = !s}")
    _start = time.perf_counter()
    blocks = list(ifeasible_subtrees(tree, root=0, M=M))
    _stop = time.perf_counter()
    print(f"Number of generated blocks: {len(blocks)}")
    print(f"Sum time: {(_stop - _start) * 1000} ms")
    mem_data = [[memory[v] for v in b] for _, b in blocks]
    blocks_data = list(zip(blocks, mem_data, map(sum, mem_data)))
    print(tabulate.tabulate(blocks_data, ('Block', 'Memory', 'Sum')))


def test_model_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml",
                        save_file: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-ser_ilp_cfg"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print(f"All unfiltered blocks: {len(list(ifeasible_subtrees(params['tree'], params['root'], math.inf)))}")
    print("  Feasible blocks  ".center(80, '='))
    for i, blk in enumerate(ifeasible_subtrees(params['tree'], params['root'], params['M'])):
        print(f"{i} ==> {blk}")
    model, _ = build_tree_cfg_model(**params)
    print("  Generated LP model  ".center(80, '='))
    # print(model)
    print_lp_desc(model)
    if save_file:
        model.writeLP("tree_cfg_model.lp")


def test_model_solution(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml"):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-ser_ilp_cfg"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    model, X = build_tree_cfg_model(**params)
    status = model.solve(solver=pulp.PULP_CBC_CMD(mip=True, msg=True))
    print("Solution:")
    pprint.pprint({x.name: x.varValue for x in model.variables()})
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    _s = time.perf_counter()
    rec_partition = recreate_subtrees_from_xdict(tree, X)
    _d = time.perf_counter() - _s
    print(f"Recreate: {_d * 1000} ms")
    _s = time.perf_counter()
    ext_partition = extract_subtrees_from_xdict(model)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {rec_partition = }, {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")


@pytest.mark.skipif(get_cplex_path() is None, reason="CPLEX is not available!")
@pytest.mark.skipif(not (sys.version_info < (3, 11)), reason="PY version is not supported!")
def test_model_solution_cplex():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  delay=10)
    print("  Run CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_hybrid_partitioning(**params, solver=pulp.CPLEX_PY(mip=True, msg=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")


@pytest.mark.skipif(get_glpk_path() is None, reason="GLPK is not available!")
def test_model_solution_glpk():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  delay=10)
    print("  Run GLPK solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_hybrid_partitioning(**params, solver=pulp.GLPK_CMD(mip=True, msg=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")


def evaluate_ilp_cfg_model():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  delay=10)
    print("  CBC solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_hybrid_partitioning(**params, solver=pulp.PULP_CBC_CMD(mip=True, msg=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)
    print("  CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_hybrid_partitioning(**params, solver=pulp.CPLEX_PY(mip=True, msg=True))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_ser_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)


########################################################################################################################


def run_test(tree: nx.DiGraph, root: int, cp_end: int, M: int, L: int, delay: int):
    partition, opt_cost, opt_lat = tree_hybrid_partitioning(tree, root, M, L, cp_end, delay,
                                                            solver=pulp.PULP_CBC_CMD(mip=True, msg=True))
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    return partition, opt_cost, opt_lat


def test_ser_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  # L = math.inf
                  L=430,
                  delay=10)
    run_test(**params)


def test_random_ser_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_feasible_subtrees(branch=2, depth=2)
    # test_restricted_feasible_subtrees()
    # test_model_creation(save_file=False)
    # test_model_solution()
    # test_model_solution_cplex()
    # test_model_solution_glpk()
    # evaluate_ilp_cfg_model()
    test_ser_tree()
    # test_random_ser_tree()
